local _dbg = require "coaf.debug"
local cjson = require "cjson"
local pl = require "pl.pretty"
local client = require "coaf.client"
local coaf = require "coaf.core"

local _M = {
    
}

_M.cpu_saved = {
	u = 0,
	n = 0,
	s = 0,
	i = 0,
	w = 0,
	x = 0,
	y = 0
}

local function GetCpuCores()
	local file = io.open( "/proc/stat", "r" )
	if not file then
		return 1
	end

	local line = file:read("l")
	local count = 0
	while true do
		line = file:read("l")
		if not line then
			break
		end
		if line:match("^cpu[0-9]+") then
			count = count + 1
		else
			break
		end
	end
	file:close()
	return count
end

local function GetCpuLoad()
	local file = io.open( "/proc/stat", "r" )
	if not file then
		return 0, 99	--load: 0, idle: 99
	end

	local line = file:read("l")
	file:close()

	local cpu,u,n,s,i,w,x,y = line:match( "(.-)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)" )
	u = u and tonumber(u) or 0
	n = n and tonumber(n) or 0
	s = s and tonumber(s) or 0
	i = i and tonumber(i) or 0
	w = w and tonumber(w) or 0
	x = x and tonumber(x) or 0
	y = y and tonumber(y) or 0

	local du = u - _M.cpu_saved.u
	local dn = n - _M.cpu_saved.n
	local ds = s - _M.cpu_saved.s
	local di = i - _M.cpu_saved.i
	local dw = w - _M.cpu_saved.w
	local dx = x - _M.cpu_saved.x
	local dy = y - _M.cpu_saved.y

	_M.cpu_saved.u = u
	_M.cpu_saved.n = n
	_M.cpu_saved.s = s
	_M.cpu_saved.i = i
	_M.cpu_saved.w = w
	_M.cpu_saved.x = x
	_M.cpu_saved.y = y

	local total = du + dn + ds + di + dw + dx + dy
	if total < 1 then
		total = 1
	end

	local scale = 100/total
	local idle = di*scale
	return 100-idle, idle
end


local function GetMemUse()
	local file = io.open( "/proc/meminfo", "r" )
	if not file then
		return nil, nil
	end

	local free = nil
	local total = nil
	local v
	local line = file:read("l")
	while line and ( not free or not total ) do
		v = line:match( "MemTotal:%s+(%d+)" )
		if v then
			total = tonumber(v)
		else
			v = line:match( "MemFree:%s+(%d+)" )
			if v then
				free = tonumber(v)
			end
		end
		line = file:read("l")
	end

	file:close()
	return free, total
end

local function GetDiskSpaces()
	local result = {
		sys = { total=0, used=0, free=0, percent=100 },
		data = { total=0, used=0, free=0, percent=100 }
	}

	local file=io.popen("/bin/df -m","r")
	if not file then
		return result
	end

	local line = file:read("l")
	while line and (result.sys.total <= 0 or result.data.total <=0 ) do
		local flag,t,u,f,p = line:match( "(.-)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%%%s+/$" )
		if flag then
			result.sys.total = tonumber(t)
			result.sys.used = tonumber(u)
			result.sys.free = tonumber(f)
			result.sys.percent = tonumber(p)
		else
			flag,t,u,f,p = line:match( "(.-)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%%%s+/data$" )
			if flag then
				result.data.total = tonumber(t)
				result.data.used = tonumber(u)
				result.data.free = tonumber(f)
				result.data.percent = tonumber(p)
			end
		end

		line=file:read("l")
	end

	file:close()
	return result
end

local function GetPersisTime()
	local file=io.open("/proc/uptime", "r")
	if not file then
		return 0, {year=1970,month=1,day=1,hour=8,min=0,sec=0}
	end

	local line = file:read("l")
	file:close()
	local pers = line:match("(%d+)")
	if not pers then
		return 0, {year=1970,month=1,day=1,hour=8,min=0,sec=0}
	else
		return tonumber(pers), os.date("*t",os.time() - tonumber(pers))
	end
end


function _M.ServerInfo()

    local cpu_c = GetCpuCores()
    local cpu_p, cpu_i = GetCpuLoad()
    local mem_f, mem_t = GetMemUse()
    local pers_m, pers_t = GetPersisTime()

    local info = {
        -- addr = ngx.var.server_addr,
        -- name = ngx.var.scheme .. "://" .. ngx.var.host,
        -- port = ngx.var.server_port,
        persis = "" .. math.floor(pers_m/3600) .. " 小时 " .. math.floor((pers_m%3600)/60) .. " 分 " .. math.floor(pers_m%60) .. "秒",
        start_time = string.format( "%04d-%02d-%02d %02d:%02d:%02d", pers_t.year, pers_t.month, pers_t.day, pers_t.hour, pers_t.min,  pers_t.sec ),
        cpu_cores = cpu_c,
        cpu_payload = math.floor(cpu_p*100)/100,
        mem_used = math.floor( ((mem_t - mem_f)/(mem_t))*100 ),
        mem_total = math.floor( (mem_t/1024)*100 ) / 100
    }

    
    return info
end


   
local function net_speed(interface, storage)
	local up, down = 0, 0

	for line in io.lines("/proc/net/dev") do

		local name = string.match(line, "^[%s]?[%s]?[%s]?[%s]?([%w]+):")
		if name == interface then
			local recv = tonumber(string.match(line, ":[%s]*([%d]+)"))
			local send = tonumber(string.match(line, "([%d]+)%s+%d+%s+%d+%s+%d+%s+%d+%s+%d+%s+%d+%s+%d$"))
			local now = os.time()

			if not storage[interface] then
				storage[interface] = { recv = 0, send = 0 ,time=now}
			else
				local interval = now - storage[interface].time
				if interval <= 0 then interval = 1 end
				down = (recv - storage[interface].recv) / interval
                up   = (send - storage[interface].send) / interval
			end

			storage[interface].time = now
			storage[interface].recv = recv
            storage[interface].send = send
            storage[interface].down = down
            storage[interface].up = up
		end
	end

    --------------------------------------------------------------------------------
	return storage
end


function _M.NetUsed()
    local data={}
    local down,up = 0
    data = net_speed("eth0", data)
    data = net_speed("eth1",data)

    coaf.Sleep(2)

    data = net_speed("eth0",data)
    data = net_speed("eth1",data)
   
    return data
end


function _M.StartDownload( down_type, url, name )
	local r, done,err
	local mod_CoManager = require "mod_CoManager"
	local device_proxy = mod_CoManager.GetProxy("/", "downloader",v)
	if not device_proxy then
		return false,"downloader server is unreached!"
	else
		r, done,err = device_proxy.StartDownload( down_type, url, name )
	end
	return done,err
end


function _M.GetDownloadStatus( down_type, name )
	local r, done,err
	local mod_CoManager = require "mod_CoManager"
	local device_proxy = mod_CoManager.GetProxy("/", "downloader",v)
	if not device_proxy then
		return false,"downloader server is unreached!"
	else
		r, done,err = device_proxy.GetDownloadStatus( down_type, name )
	end
	return done,err
end

--代表设备重启过了，置为 0 ------同步更新api 调用
--_M.flag=firmware 固件下载ok


_M.flag = 0

function _M.GetOperationFlag()
	return {flag = _M.flag}
end

function _M.UpdateOperationFlag(type)

	_M.flag = type
	_dbg.Info("DEVICE-MANAGER", "UpdateOperationFlag=_M.flag=", _M.flag)

	return true
end

return _M


