local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
require "kmCore"

local _M = {}

---------------------------------

local tap_meta = {}

function tap_meta:Init()
	if self.dispatcher then --Already initialized.
		return true
	end

	self.dispatcher = KM_STREAMING.AudioEngine.CreateAudioDispatcher( self.name )
	if not self.dispatcher then
		_dbg.Error( "MIXER-TAP", "Fail to create audio dispatcher for tap:", self.name )
		return false
	end

	return true
end

function tap_meta:SetOwnerMixer(mixer, isMixOutputName)
	self.mixer = mixer
	self.isMixOutputName = isMixOutputName
end

function tap_meta:Destroy()
	self:FreeEncodingChannel(nil,true)	-- Free all.
	self:RemoveInput()
	self:RemoveAllOutputs(true) -- include mix output.
	collectgarbage()
end

function tap_meta:ConnectToMixOutput(mixOutName)
	if not self.mixer then
		_dbg.Error( "MIXER-TAP", "ConnectToMixOutput: Invalid mixer bound for:", self.name )
		return false
	end
	if not mixOutName then
		_dbg.Error( "MIXER-TAP", "ConnectToMixOutput: Invalid mix-output name for:", self.name )
		return false
	end
	local MIXER_DRV = self.mixer.MIXER_DRV
	local allow, conn_type = MIXER_DRV.CheckMixerConnection( self.name, mixOutName )
	if not allow then
		_dbg.Warn( "MIXER-TAP", "ConnectToMixOutput: Mixer driver does not allow connect to mixer:", self.name, "->", mixOutName )
		return false
	end

	if conn_type == "#LOOP-DIRECTLY#" then
		local mixer_tap = self.mixer:GetMixOutputTap(mixOutName)
		if not mixer_tap then
			_dbg.Error( "MIXER-TAP", "ConnectToMixOutput: Try directly connect to mix output but fail to get mix-output tap:", mixOutName, "for:", self.name )
			return false
		end
		local disp = mixer_tap:GetAudioDispatcher()
		if not disp then
			_dbg.Error( "MIXER-TAP", "ConnectToMixOutput: Try directly connect to mix output but the tap without audio dispatcher:", mixOutName, "for:", self.name )
			return false
		end

		if self:AddOutputSink( ".mix-out:" .. mixOutName, disp ) then
			self.connectedMixOutputs[mixOutName] = true
			return true
		else
			return false
		end
	end

	local target, args = MIXER_DRV.GetMixerSinkTypeAndArgs( mixOutName )
	if target then
		if self:AddOutputTarget( ".mix-out:" .. mixOutName, target, args ) then
			self.connectedMixOutputs[mixOutName] = true
			return true
		else
			return false
		end
	else
		return false --true
	end
end

function tap_meta:DisconnectFromMixOutput(mixOutName)
	if not self.mixer then
		_dbg.Error( "MIXER-TAP", "DisconnectFromMixOutput: Invalid mixer bound for:", self.name )
		return false
	end
	if not mixOutName then
		_dbg.Error( "MIXER-TAP", "DisconnectFromMixOutput: Invalid mix-output name for:", self.name )
		return false
	end
	self:FreeEncodingChannelsByMixOutput(mixOutName,true)
	if self:RemoveOutput( ".mix-out:" .. mixOutName ) then
		self.mixer.MIXER_DRV.ReleaseMixerConnection( self.name, mixOutName )
		self.connectedMixOutputs[mixOutName] = nil
		return true
	else
		return false
	end
end

function tap_meta:AddOutputTarget( name, sink_type, sink_args )
	if not self.dispatcher then
		_dbg.Error( "MIXER-TAP", "AddOutputTarget: You must initialize tap first! For:", self.name )
		return false
	end
	if not name or not sink_type then
		_dbg.Error( "MIXER-TAP", "AddOutputTarget: Invalid output name or sink type! For:", self.name )
		return false
	end

	if self.outputs[name] then
		_dbg.Warn( "MIXER-TAP", "AddOutputTarget: The output target '" .. name .. "' is already exist! For:", self.name )
		return true
	end

	local sinkObj = KM_STREAMING.AudioEngine.CreateAudioSink( sink_type, sink_args )
	if not sinkObj then
		_dbg.Error( "MIXER-TAP", "AddOutputTarget: Fail to create output sink - type:", sink_type, " for:", self.name )
		return false
	end

	local id = self.dispatcher:AddOutputSink(sinkObj)
	if not id or id < 0 then
		_dbg.Error( "MIXER-TAP", "AddOutputTarget: Fail to add output sink to dispatcher - type:", sink_type, " for:", self.name )
		KM_STREAMING.AudioEngine.FreeAudioSink( sinkObj )
		return false
	end

	--Added output sink, so require mix output tap source for working.
	if self.isMixOutputName and self.mixer then
		self.mixer:_RequireMixOutputTapSource(self.isMixOutputName)
	end

	--Added output sink, so require default tap source for working.
	self.outputs[name] = { internal = true, obj = sinkObj, sink_type = sink_type, sink_args = sink_args }
	return true
end

function tap_meta:AddOutputSink( name, sinkObj )
	if not self.dispatcher then
		_dbg.Error( "MIXER-TAP", "AddOutputSink: You must initialize tap first! For:", self.name )
		return false
	end
	if not name or not sinkObj then
		_dbg.Error( "MIXER-TAP", "AddOutputSink: Invalid output name or sink object! For:", self.name )
		return false
	end

	if self.outputs[name] then
		_dbg.Warn( "MIXER-TAP", "AddOutputSink: The output target '" .. name .. "' is already exist! For:", self.name )
		return true
	end

	local id = self.dispatcher:AddOutputSink(sinkObj)
	if not id or id < 0 then
		_dbg.Error( "MIXER-TAP", "AddOutputSink: Fail to add output sink to dispatcher - sink object name:", name, " for:", self.name )
		return false
	end

	--Added output sink, so require mix output tap source for working.
	if self.isMixOutputName and self.mixer then
		self.mixer:_RequireMixOutputTapSource(self.isMixOutputName)
	end

	self.outputs[name] = { internal = false, obj = sinkObj }
	return true
end

--Internal using.
function tap_meta:_CheckRemoveMixOutputSource()
	if self.isMixOutputName and self.mixer then
		local k,v
		local remained_outputs
		for k,v in pairs(self.outputs) do
			remained_outputs = true
			break
		end
		if not remained_outputs then
			self.mixer:_ReleaseMixOutputTapSource(self.isMixOutputName)
		end
	end
end

function tap_meta:RemoveOutput( name )
	if not self.dispatcher then
		_dbg.Error( "MIXER-TAP", "RemoveOutput: You must initialize tap first! For:", self.name )
		return false
	end
	if not name then
		_dbg.Error( "MIXER-TAP", "RemoveOutput: Invalid output name! For:", self.name )
		return false
	end

	local out_item = self.outputs[name]
	if not out_item then
		return true
	end

	if out_item.obj then
		self.dispatcher:RemoveOutputSink(out_item.obj)
	end

	if out_item.internal then
		KM_STREAMING.AudioEngine.FreeAudioSink( out_item.obj )
	end

	self.outputs[name] = nil

	self:_CheckRemoveMixOutputSource()
	return true
end

function tap_meta:RemoveAllOutputs(includeMixOutput)
	if not self.dispatcher then
		_dbg.Error( "MIXER-TAP", "RemoveAllOutputs: You must initialize tap first! For:", self.name )
		return false
	end

	local k,out_item
	for k,out_item in pairs(self.outputs) do
		if includeMixOutput or not k:match( "%.mix%-out:" ) then
			if out_item.obj then
				self.dispatcher:RemoveOutputSink(out_item.obj)
			end
		
			if out_item.internal then
				KM_STREAMING.AudioEngine.FreeAudioSink( out_item.obj )
			end
		
			self.outputs[k] = nil
		end
	end

	self:_CheckRemoveMixOutputSource()
	return true
end

function tap_meta:SelectInputSource( source_type, source_args )
	if not self.dispatcher then
		_dbg.Error( "MIXER-TAP", "SelectInputSource: You must initialize tap first! For:", self.name )
		return false
	end

	if not source_type then
		_dbg.Error( "MIXER-TAP", "SelectInputSource: Invalid input source type specified! For:", self.name )
		return false
	end

	local sourceObj = KM_STREAMING.AudioEngine.CreateAudioSource( source_type, source_args )
	if not sourceObj then
		_dbg.Error( "MIXER-TAP", "SelectInputSource: Fail to create audio source of type:", source_type, "For:", self.name )
		return false
	end

	if self.input then
		self:RemoveInput()
	end

	if not self.dispatcher:SelectAudioSourceAsInput( sourceObj ) then
		_dbg.Error( "MIXER-TAP", "SelectInputSource: Fail to select audio source as input for dispatcher, type:", source_type, "For:", self.name )
		KM_STREAMING.AudioEngine.FreeAudioSource( sourceObj )
		return false
	end

	self.input = {
		internal = true,
		obj = sourceObj,
		source_type = source_type,
		source_args = source_args
	}

	return true
end

function tap_meta:SelectInputSourceObject( sourceObj )
	if not self.dispatcher then
		_dbg.Error( "MIXER-TAP", "SelectInputSourceObject: You must initialize tap first! For:", self.name )
		return false
	end

	if not sourceObj then
		_dbg.Error( "MIXER-TAP", "SelectInputSourceObject: Invalid input source object specified! For:", self.name )
		return false
	end

	if self.input then
		self:RemoveInput()
	end

	if not self.dispatcher:SelectAudioSourceAsInput( sourceObj ) then
		_dbg.Error( "MIXER-TAP", "SelectInputSourceObject: Fail to select audio source object as input for dispatcher, For:", self.name )
		return false
	end

	self.input = {
		internal = false,
		obj = sourceObj
	}

	return true
end

function tap_meta:SelectInputSink( audioSink )
	if not self.dispatcher then
		_dbg.Error( "MIXER-TAP", "SelectInputSink: You must initialize tap first! For:", self.name )
		return false
	end

	if not audioSink then
		_dbg.Error( "MIXER-TAP", "SelectInputSink: Invalid input sink object specified! For:", self.name )
		return false
	end

	if self.input then
		self:RemoveInput()
	end

	if not self.dispatcher:SelectAudioSinkAsInput( audioSink ) then
		_dbg.Error( "MIXER-TAP", "SelectInputSink: Fail to select audio sink object as input for dispatcher, For:", self.name )
		return false
	end

	self.input = {
		internal = false,
		obj = audioSink
	}

	return true
end

function tap_meta:RemoveInput()
	if not self.dispatcher then
		_dbg.Error( "MIXER-TAP", "RemoveInput: You must initialize tap first! For:", self.name )
		return false
	end
	if self.input then
		self.dispatcher:RemoveInput()
		if self.internal and self.obj then
			KM_STREAMING.AudioEngine.FreeAudioSource( self.obj )
		end
		self.input = nil
	end
	return true
end

function tap_meta:GetAudioDispatcher()
	return self.dispatcher
end

function tap_meta:Enable()
	if not self.dispatcher then
		_dbg.Error( "MIXER-TAP", "Enable: You must initialize tap first! For:", self.name )
		return false
	end
	return self.dispatcher:Open()
end

function tap_meta:Disable()
	if not self.dispatcher then
		_dbg.Error( "MIXER-TAP", "Disable: You must initialize tap first! For:", self.name )
		return false
	end
	return self.dispatcher:Close()
end

function tap_meta:CreateEncodingChannel(codecArgs,mixOutName)
	if not codecArgs then
		_dbg.Error( "MIXER-TAP", "CreateEncodingChannel: Invalid encoding args specified! For:", self.name )
		return nil
	end
	if not self.dispatcher then
		_dbg.Error( "MIXER-TAP", "CreateEncodingChannel: You must initialize tap first! For:", self.name )
		return nil
	end

	if not self.mixer then
		_dbg.Error( "MIXER-TAP", "CreateEncodingChannel: No owner mixer bond for tap:", self.name )
		return nil
	end

	if not codecArgs.type or
	   not codecArgs.channels or
	   not codecArgs.sampling or
	   not codecArgs.sampleSize then
		_dbg.Error( "MIXER-TAP", "CreateEncodingChannel: Invalid encoding args(type,channels,sampling,...) for tap:", self.name )
		return nil
	end

	mixOutName = mixOutName or self.isMixOutputName --Using default mix-output, if bond.
	if not mixOutName then
		_dbg.Error( "MIXER-TAP", "CreateEncodingChannel: No valid mix-output name specified for tap:", self.name )
		return nil
	end

	local argsIdr = "MX:" .. mixOutName .. ",T:" .. codecArgs.type .. ",C:" .. codecArgs.channels .. ",SMP:" .. codecArgs.sampling .. ",SZ:" .. codecArgs.sampleSize
	if codecArgs.ID then
		argsIdr = argsIdr .. ",ID:" .. codecArgs.ID
	end
	if codecArgs.type == "AAC" or codecArgs.type == "AAC-LC" or codecArgs.type == "MPEG4-GENERIC" then
		argsIdr = argsIdr .. ",SP:"
		if codecArgs.aacType then
			argsIdr = argsIdr .. codecArgs.aacType
		end
		if codecArgs.aacFormat then
			argsIdr = argsIdr .. " " .. codecArgs.aacFormat
		end
	elseif codecArgs.type == "G711" or codecArgs.type == "PCMU" or codecArgs.type == "PCMA" 
		   or codecArgs.type == "G711A" or codecArgs.type == "G711U"
		   or codecArgs.type == "G.711" or codecArgs.type == "G.711A" or codecArgs.type == "G.711U" then
		argsIdr = argsIdr .. ",SP:" .. codecArgs.g711Law
	end

	if self.encodingChannels[ argsIdr ] then
		_dbg.Info( "MIXER-TAP", "CreateEncodingChannel: Using created encoding channel:", argsIdr, " for tap:", self.name ) 
		self.encodingChannels[ argsIdr ].ref_count = self.encodingChannels[ argsIdr ].ref_count + 1
		return self.encodingChannels[ argsIdr ].obj, argsIdr
	end

	if not _M.audioEncEng then
		_M.audioEncEng = KM_STREAMING.AudioEngine.Encoder.createNew()
		if not _M.audioEncEng then
			_dbg.Error( "MIXER-TAP", "CreateEncodingChannel: Fail to initialize audio encoding engine for tap:", self.name )
			return nil
		end
	end

	--TODO: I think no need to restrict this.
	--if mixOutName and not self.connectedMixOutputs[mixOutName] then
	--	_dbg.Error( "MIXER-TAP", "CreateEncodingChannel: Current tap is not connect to mix-output:", mixOutName, "For tap:", self.name )
	--	return nil
	--end
	
	if not self.encoder_mixers[mixOutName] then
		local srcArgs, err = self.mixer:_CreateEncoderMixer( self.name, mixOutName, self.isMixOutputName and true or false )
		if not srcArgs then
			_dbg.Warn( "MIXER-TAP", "CreateEncodingChannel: Fail to create special encoder mixer for tap:", self.name, "Error:", err, " so try to retrival the default mixer for encoding!" )
		else
			self.encoder_mixers[mixOutName] = srcArgs
		end
	end

	local enc_chn = nil
	local err
	if not self.encoder_mixers[mixOutName] then
		--Try create encoding channel using default mixer.
		enc_chn, err = self.mixer:_CreateDefaultEncodingChannel(mixOutName, codecArgs, self.name)
		if not enc_chn then
			_dbg.Error( "MIXER-TAP", "CreateEncodingChannel: Fail to create encoding channel using default mixer for tap:", self.name, "Error:", err )
			return nil
		else
			_dbg.Info( "MIXER-TAP", "CreateEncodingChannel: Created encoding channel for tap:", self.name, "using default mixer!" )
		end
	else
		enc_chn = _M.audioEncEng:CreateEncodingChannel( self.encoder_mixers[mixOutName], codecArgs )
		if not enc_chn then
			_dbg.Error( "MIXER-TAP", "CreateEncodingChannel: Fail to create audio encoding channel for tap:", self.name )
			return nil
		else
			_dbg.Info( "MIXER-TAP", "CreateEncodingChannel: Created encoding channel for tap:", self.name, "using mixer:", self.encoder_mixers[mixOutName] )
			enc_chn:SetParameters(codecArgs)
		end
	end

	self.encodingChannels[argsIdr] = {
		args = codecArgs,
		encoder_mixer = self.encoder_mixers[mixOutName] and self.encoder_mixers[mixOutName].mixer_name or nil,
		mix_out_name = mixOutName,
		obj = enc_chn,
		ref_count = 1
	}

	return enc_chn, argsIdr
end

function tap_meta:FreeEncodingChannelsByMixOutput(mixOutName,force)
	local k, item
	local removes = {}
	local should_remove_mixers = {}
	for k, item in pairs(self.encodingChannels) do
		if item.mix_out_name == mixOutName then
			if force then
				item.ref_count = 0
			elseif item.ref_count > 0 then
				item.ref_count = item.ref_count - 1
			end

			if item.ref_count <= 0 then
				if item.encoder_mixer then
					should_remove_mixers[ item.encoder_mixer ] = mixOutName
				end

				if item.encoder_mixer then
					item.obj:Destroy()
				else
					self.mixer:_FreeDefaultEncodingChannel(item.obj, item.mix_out_name)
				end
				item.obj = nil
				table.insert( removes, k )
			end
		end
	end

	for k, item in pairs(removes) do
		self.encodingChannels[item] = nil
	end

	for k, item in pairs(self.encodingChannels) do
		--If the mixer is using yet, don't remove it.
		if item.encoder_mixer and should_remove_mixers[ item.encoder_mixer ] then
			should_remove_mixers[ item.encoder_mixer ] = nil
		end
	end

	for k, item in pairs(should_remove_mixers) do
		self.mixer:_FreeEncoderMixer( self.name, item, k )
		self.encoder_mixers[item] = nil
	end
end

function tap_meta:FreeEncodingChannel(chan,force)
	local k, item
	local removes = {}
	local should_remove_mixers = {}
	for k, item in pairs(self.encodingChannels) do
		if not chan or item.obj == chan then
			if force then
				item.ref_count = 0
			elseif item.ref_count > 0 then
				item.ref_count = item.ref_count - 1
			end

			if item.ref_count <= 0 then
				if item.encoder_mixer then
					should_remove_mixers[ item.encoder_mixer ] = item.mix_out_name
				end

				if item.encoder_mixer then
					item.obj:Destroy()
				else
					self.mixer:_FreeDefaultEncodingChannel(item.obj, item.mix_out_name)
				end
				item.obj = nil
				table.insert( removes, k )
			end

			if chan then
				break
			end
		end
	end

	for k, item in pairs(removes) do
		self.encodingChannels[item] = nil
	end

	for k, item in pairs(self.encodingChannels) do
		--If the mixer is using yet, don't remove it.
		if item.encoder_mixer and should_remove_mixers[ item.encoder_mixer ] then
			should_remove_mixers[ item.encoder_mixer ] = nil
		end
	end

	for k, item in pairs(should_remove_mixers) do
		self.mixer:_FreeEncoderMixer( self.name, item, k )
		self.encoder_mixers[item] = nil
	end
end

tap_meta.__index = tap_meta

--------------------------------

local default_id_counter = 0

function _M.Create(name)
	if not name then
		name = "tap" .. default_id_counter
		default_id_counter = default_id_counter + 1
	end

	return setmetatable({
		name = name,
		mixer = nil,
		isMixOutputName = nil,
		input = nil,
		input_obj = nil,
		outputs = {},
		encodingChannels = {},
		connectedMixOutputs = {},
		encoder_mixers = {}
	}, tap_meta )
end

return _M

