local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"

require "kmCore"
local sys_audio_config = require "sys_audio_config"

_G.audio_mix_config = require "configs.audio_mix_config".Load()

local _done, features = pcall(require, "features")
if not _done then
	features = {}
elseif type(features) ~= "table" then
	features = _G
end

local _M = nil

if features.FEATURE_USB_SOUND_DEVEICE then
	_M = {
		channel = nil,
		usb_channel = nil,
		vs0_channel = nil,
		vs1_channel = nil,
		audio_cfg = {
				usb_channel = {mute = _G.audio_mix_config.usb_channel.mute, volume = _G.audio_mix_config.usb_channel.volume}
		},

		usb_probe_thread = nil
	}
else
	_M = {
		channel = nil,
		usb_channel = nil,
		vs0_channel = nil,
		vs1_channel = nil,
		audio_cfg = {
		}
	}
end

local function file_line( filename )
	local f = io.open( filename, "r" )
	if f then
		local r = f:read("*l")
		f:close()
		return r
	else
		return nil
	end
end

local function check_usb_sound_device(ignore_device)
	local item
	local card_id, card_name, card_fnd
	local lfs = require "lfs"
	for item in lfs.dir("/sys/class/sound") do
		if item:match("card%d+") then
			card_id = tonumber(file_line("/sys/class/sound/" .. item .. "/number"))
			card_name = file_line("/sys/class/sound/" .. item .. "/id")
			if card_id and card_name and ignore_device and ignore_device ~= "" and ignore_device == card_name then
				card_name = nil
			end

			if card_name then -- Check is USB device?
				if tonumber( "0x" .. (file_line("/sys/class/sound/" .. item .. "/device/bInterfaceClass") or "#")) then
					card_fnd = card_id
					break
				end
			end
		end
	end
	return card_fnd
end

local function check_sound_bus_id(card_id)
	local f = io.popen("/usr/bin/readlink -f /sys/class/sound/card" .. card_id, "r" )
	if f then
		local path = f:read("*l")
		f:close()
		if not path then
			return nil, nil
		end

		local dev_path = path:match("^(.*/usb[0-9]+/[0-9]+%-[0-9]+)/.*")
		if not dev_path then
			return nil, nil
		end

		return tonumber(file_line(dev_path .. "/busnum")), tonumber(file_line(dev_path .. "/devnum"))
	else
		return nil, nil
	end
end

local function detect_usb_audio(throbj, ignore_device, bind_output)
	local mod_AudioOut = require "mod_AudioOut"

	_dbg.Info( "AUDIO_ENC", "USB audio probing task is running...")

	local card_fnd = nil

	while not throbj.terminated do
		if not card_fnd then --Check the USB sound card
			card_fnd = check_usb_sound_device(ignore_device)
			if card_fnd then -- Update the USB sound ALSA source.
				local new_src = KM_STREAMING.AudioEngine.CreateAudioSource( "ALSA", {
					device = "hw:" .. card_fnd,
					fixed_channels = 1 --USB audio, now force be 1 channel.
				})
				if new_src and _M.usb_channel then
					_M.usb_channel:ChangeSource( new_src )
				end

				if bind_output then
					mod_AudioOut.AddOutputSink( bind_output, "USB", "ALSA", {device = "hw:" .. card_fnd} )
					local dispatcher = mod_AudioOut.outputs[sys_audio_config.usb.bind_output].mix_out_tap.dispatcher
					if dispatcher then
						local zero = dispatcher:SetGainRange(200, -51, 20)
						if _M.audio_cfg.usb_channel then
							if _G.audio_mix_config.usb_channel.mute == 1 then
								dispatcher:SetGainValue(0, 0)
							else
								dispatcher:SetGainValue(_G.audio_mix_config.usb_channel.volume, _G.audio_mix_config.usb_channel.volume)
							end
						else
							local volume = _G.audio_mix_config.usb_channel.volume
							dispatcher:SetGainValue(volume, volume)
							_M.audio_cfg.usb_channel = {volume = volume, mute = _G.audio_mix_config.usb_channel.mute, zero = zero, device = _M.usb_channel}
						end
					end
				end
				--TODO: Add audio mixer.
			end
		else -- Check USB sound card still exist.
			if not tonumber("0x" .. (file_line("/sys/class/sound/card" .. card_fnd .. "/device/bInterfaceClass") or "#")) then
				card_fnd = nil
				if _M.usb_channel then
					_M.usb_channel:ChangeSource(nil)
				end
				if bind_output then
					mod_AudioOut.RemoveOutputSink( bind_output, "USB" )
				end
			end
		end
		coaf.Sleep(1)
	end
	_dbg.Info( "AUDIO_ENC", "USB audio probing task is terminated!")
end

function _M.Init()
	local outputs = require "mod_Output".all_outputs
	local mod_AudioOut = require "mod_AudioOut"

	_M.audioEng = KM_STREAMING.AudioEngine.Encoder.createNew()
	if not _M.audioEng then
		_dbg.Error( "AUDIO_ENC", "Fail to create audio engine!" )
	else
		_M.channel = _M.audioEng:CreateEncodingChannel( { type="HISIAI", device="default" }, { type="AAC", bufferSize = 512000, channels=2 } )
		if not _M.channel then
			_dbg.Error( "AUDIO_ENC", "Fail to create audio encoding channel!" )
		else
			local mo_ain = require "mod_AudioIn"
			if mo_ain.ainChannel then
				_M.channel:SetParameters({
					sampleSize = mo_ain.ainChannel.aiPubAttr.u32PtNumPerFrm,
					channels = mo_ain.ainChannel.aiPubAttr.u32ChnCnt,
					sampling = mo_ain.ainChannel.aiPubAttr.enSamplerate,
					aacType = "AAC-LC",
					aacFormat = "RAW",
					bitrate = 64000
				})
			end

			local cfg = _G.audio_mix_config.channel
			local volume = cfg.volume

			_M.audio_cfg.channel = {mute = cfg.mute, volume = volume, zero = cfg.zero, device = _M.channel}
			local mod_AudioOut = require "mod_AudioOut"

			if features.FEATURE_USB_SOUND_DEVEICE then
				mod_AudioOut.SetOutputVolume(sys_audio_config.usb.bind_output, volume )
				mod_AudioOut.SetOutputVolume(sys_audio_config.usb.bind_output, cfg.mute == 1 and "mute" or "unmute" )
			else
				mod_AudioOut.SetOutputVolume("LINEOUT", volume )
				mod_AudioOut.SetOutputVolume("LINEOUT", cfg.mute == 1 and "mute" or "unmute" )
			end
		end

		if features.FEATURE_USB_SOUND_DEVEICE then
			if sys_audio_config.usb and sys_audio_config.usb.enable then
				--NOTE: This is the USB audio driver bug, so I *MUST* reset the USB bus before using it.
				local card = check_usb_sound_device(sys_audio_config.usb.ignore)
				if card then
					local busnum, devnum = check_sound_bus_id(card)
					if busnum and devnum then
						os.execute("/usr/bin/usbreset /dev/bus/usb/" .. string.format("%03d", busnum) .. "/" .. string.format("%03d", devnum))
					end
				end

				-- Since I can't ensure which USB device current is, so initialize a NONE audio source.
				_M.usb_channel = _M.audioEng:CreateEncodingChannel( { type="NONE" }, { type="AAC", bufferSize = 512000, channels=2 } )
				if not _M.usb_channel then
					_dbg.Error( "AUDIO_ENC", "Fail to create USB audio encoding channel!" )
				else
					_M.usb_channel:SetParameters({
						sampleSize = 1024,
						channels = 2,
						sampling = 48000,
						aacType = "AAC-LC",
						aacFormat = "RAW",
						bitrate = 64000
					})

					_M.usb_probe_thread = thread.New( detect_usb_audio )
					if _M.usb_probe_thread then
						_M.usb_probe_thread:Start( sys_audio_config.usb.ignore, sys_audio_config.usb.bind_output )
					end
				end
			end
		end

		if features.FEATURE_AUDIO_MIX_OUTPUT then
			_M.vs0_channel = _M.audioEng:CreateEncodingChannel( {type="CROSS"}, {type="AAC", bufferSize=512000, channels=2} )
			if not _M.vs0_channel then
				_dbg.Error( "AUDIO_ENC", "Fail to create virtual screen audio encoding channel (0)!")
			else
				_M.vs0_channel:SetParameters({
					sampleSize = 1024,
					channels = 2,
					sampling = 48000,
					aacType = "AAC-LC",
					aacFormat = "RAW",
					bitrate = 64000
				})
				local source = _M.vs0_channel:GetSource()
				if source then
					local zero = source:SetGainRange(200, -51, 20)
					local volume = _G.audio_mix_config.vs0_channel.volume
					source:SetGainValue(volume, volume)
					local mute = _G.audio_mix_config.vs0_channel.mute
					if mute == 1 then
						source:SetGainValue(0, 0)
						--source:Mute(true)
					end
					_M.audio_cfg.vs0_channel = {mute = mute, volume = volume, zero = zero, device = _M.vs0_channel}
					local sink = mod_AudioOut.GetOutputSink("VS0")
					if sink then
						KM_STREAMING.AudioEngine.ConnectSinkToSource(sink, source);
					end
				end
			end

			_M.vs1_channel = _M.audioEng:CreateEncodingChannel( {type="CROSS"}, {type="AAC", bufferSize=512000, channels=2} )
			if not _M.vs1_channel then
				_dbg.Error( "AUDIO_ENC", "Fail to create virtual screen audio encoding channel (1)!")
			else
				_M.vs1_channel:SetParameters({
					sampleSize = 1024,
					channels = 2,
					sampling = 48000,
					aacType = "AAC-LC",
					aacFormat = "RAW",
					bitrate = 64000
				})
				local source = _M.vs1_channel:GetSource()
				if source then
					local zero = source:SetGainRange(200, -51, 20)
					local volume = _G.audio_mix_config.vs1_channel.volume
					source:SetGainValue(volume, volume)
					local mute = _G.audio_mix_config.vs1_channel.mute
					if mute == 1 then
						source:SetGainValue(0, 0)
						--source:Mute(true)
					end
					_M.audio_cfg.vs1_channel = {mute = mute, volume = volume, zero = zero, device = _M.vs1_channel}
					local sink = mod_AudioOut.GetOutputSink("VS1")
					if sink then
						KM_STREAMING.AudioEngine.ConnectSinkToSource(sink, source);
					end
				end
			end
		end
	end

	for name, output in pairs(outputs) do
		_G.audio_mix_config[name] = _G.audio_mix_config[name] or {}
		if _G.audio_mix_config[name].mute and _G.audio_mix_config[name].mute == 1 then
			for i,v in ipairs(output.output_config.bind_audio_outputs) do
				if sys_audio_config.outputs and sys_audio_config.outputs[v] and sys_audio_config.outputs[v].audio_type == "embedded" then
					mod_AudioOut.SetOutputVolume( v, "mute" )
				end
			end
		end
	end
end

function _M.GetAudioInfo()
	local info = {}
	for name, obj in pairs(_M.audio_cfg) do
		local tbl = {mute = obj.mute, volume = obj.volume, name = name, device = name}
		-- if name == "channel" then
		-- 	tbl.localoutput = obj.localoutput
		-- end
		table.insert(info, tbl)
	end
	return info
end

function _M.ModifyMuteEmbed(outname, mute)
	_G.audio_mix_config[outname].mute = mute and 1 or 0
	_G.audio_mix_config:__TRIGGER(true)
	local outputs = require "mod_Output".all_outputs
	local mod_AudioOut = require "mod_AudioOut"
	local output = outputs[outname]
	if output then
		for i,v in ipairs(output.output_config.bind_audio_outputs) do
			if sys_audio_config.outputs and sys_audio_config.outputs[v] and sys_audio_config.outputs[v].audio_type == "embedded" then
				mod_AudioOut.SetOutputVolume( v, mute and "mute" or "unmute" )
			end
		end
	end
	return true
end

function _M.GetEmbedMute(outname)
	return _G.audio_mix_config[outname] and _G.audio_mix_config[outname].mute or 0
end

function _M.ModifyAduioInfo(device, arg)
	local obj = _M.audio_cfg[device]
	if not obj then
		return false, "Invalid device"
	end

	if arg.mute then
		obj.mute = arg.mute
		_G.audio_mix_config[device].mute = tonumber(arg.mute)
		_G.audio_mix_config:__TRIGGER(true)
	end
	if arg.volume then
		obj.volume = arg.volume
		_G.audio_mix_config[device].volume = tonumber(arg.volume)
		_G.audio_mix_config:__TRIGGER(true)
	end
	if arg.localoutput then
		obj.localoutput = arg.localoutput
	end

	if device == "channel" then
		if arg.localoutput then
			local outputs = require "mod_Output".all_outputs
			for _, v in pairs(outputs) do
				v.SetAudioGain("analog", "none", tonumber(obj.localoutput))
			end
			obj.localoutput = nil
		else
			local mod_AudioOut = require "mod_AudioOut"
			local vol 
			if arg.mute then
				vol = arg.mute == 1 and "mute" or "unmute"
			elseif arg.volume then
				vol = tonumber(obj.volume)
			end
			if vol then
				if features.FEATURE_USB_SOUND_DEVEICE then
					mod_AudioOut.SetOutputVolume(sys_audio_config.usb.bind_output, vol )
				else
					mod_AudioOut.SetOutputVolume("LINEOUT", vol )
				end
			end
		end
	
		return true
	end
	
	local source = (obj.device or _M[device]):GetSource()
	if source then
		if obj.mute == 1 then
			source:SetGainValue(0, 0)
		else
			source:SetGainRange(200, -51, 20)
			source:SetGainValue(obj.volume, obj.volume)
		end
	else
		_dbg.Error("AudioEncoder", "Do not have source.", device)
		-- local mod_AudioOut = require "mod_AudioOut"
		-- local out = mod_AudioOut.outputs[sys_audio_config.usb.bind_output]
		-- local dispatcher = out and out.mix_out_tap and out.mix_out_tap.dispatcher
		-- if not dispatcher then
		-- 	return false
		-- end
		-- if obj.mute == 1 then
		-- 	dispatcher:Mute(true)
		-- else
		-- 	dispatcher:Mute(false)
		-- 	dispatcher:SetGainValue(obj.volume, obj.volume)
		-- end

		-- if obj.localoutput and device == "channel" then
		-- 	local outputs = require "mod_Output".all_outputs
		-- 	for _, v in pairs(outputs) do
		-- 		v.SetAudioGain("analog", "none", tonumber(obj.localoutput))
		-- 	end
		-- end
	end
	return true
end

return _M

