local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local client = require "coaf.client"
local posix_time = require "posix.time"
local streams = require "mod_Streams"
local pt = require "pl.pretty"

local function GetMonoTime()
	local tm = posix_time.clock_gettime( posix_time.CLOCK_MONOTONIC )
	if tm then
		return tm.tv_sec
	else
		return os.time()
	end
end

local _M = {
	devices = {},
	coserver_client = nil,
	coserver_proxy = nil,
	module_alives = {},
	serialno = "",
	ui_client = nil,
	ui_proxy = nil,
	synctime_dev = {}
}

local function GetProxy( obj, svc, loc, timeout )
	local proxy
	local cli
	cli = client.New( svc, loc, nil, timeout or 3 ) --3 seconds timeout.
	if not cli then
		_dbg.Error( "DEVICE_MANAGER", "Fail to create connection to service " .. (loc or "local") .. "/" .. (svc or "nil") )
		return nil
	end
	proxy = cli:CreateProxy(obj, true) --Exclusive
	if not proxy then
		_dbg.Error( "DEVICE_MANAGER", "Fail to create object proxy for '" .. (obj or "unknown") .. "', service " .. (loc or "local") .. "/" .. (svc or "nil") )		
		return nil
	end
	return proxy
end

--更新远程设备的设备状态。
--TODO: 从设备CoAF服务上获取状态 (NOTE by左振宇)

function _M.UpdateDeviceStatus( device_id )
	local detail = _M.devices[ device_id ]
	if not detail then
		return false
	end

	-- if detail.address then
	-- 	detail.url = "rtsp://" .. detail.address .. "/main"
	-- 	detail.url2 = "rtsp://" .. detail.address .. "/second"
	-- end

	--TODO: Check more details.
	--local proxy = GetProxy( "/", "systemctrl", device_id )
	--if not proxy then
	--	return false
	--end

	--proxy:Destroy()

	return true
end


function _M.GetDeviceDetail( device_id )
	return _M.devices[ device_id ]
end

function _M.GetDeviceList()
	return _M.devices
end

function _M.GetDeviceIDList()
	local rslt = {}
	local k,v
	for k,v in pairs(_M.devices) do
		table.insert( rslt, k )
	end
	table.sort( rslt, function(a,b)
		if not a or not b or a == b then
			return false
		else
			return a < b
		end
	end)
	return rslt
end

local function CheckDeviceOnline(id)
--[[
	if not _M.coserver_proxy then
		return false
	end

	local r, ret = _M.coserver_proxy.GetLocations(id,"online")
	if not r or type(ret) ~= "table" then
		return false
	end

	if ret[id] == "online" then
		return true
	else
		return false
	end
--]]
	if not id or not _M.devices[id] then
		return false
	else
		return _M.devices[id].status == "on"
	end
end

function _M.ResetDevice(id)
	if not CheckDeviceOnline(id) then
		return false,"模块不在线"
	end
	local proxy = GetProxy( "/", "systemctrl", id )
	if not proxy then
		return false, "无法连接模块控制服务程序"
	end
	proxy.Reset()
	proxy:Destroy()
	return true
end

function _M.RebootDevice(id)
	if not CheckDeviceOnline(id) then
		return false,"模块不在线"
	end

	local info = _M.devices[id]
	if info then
		local now = GetMonoTime()
		info.keep_status_util = now + 20
	end

	local proxy = GetProxy( "/", "systemctrl", id )
	if not proxy then
		if info then
			info.work = "reboot-fail"
		end
		return false, "无法连接模块控制服务程序"
	end
	proxy.Reboot(2) --Delay 2s
	proxy:Destroy()

	if info then
		info.work = "reboot"
	end

	return true
end

function _M.RestoreDevice(id)
	if not CheckDeviceOnline(id) then
		return false,"模块不在线"
	end

	local info = _M.devices[id]
	if info then
		local now = GetMonoTime()
		info.keep_status_util = now + 20
	end

	local proxy = GetProxy( "/", "systemctrl", id )
	if not proxy then
		if info then
			info.work = "restore-fail"
		end
		return false, "无法连接模块控制服务程序"
	end
	proxy.RestoreFactory()
	proxy:Destroy()

	if info then
		info.work = "restore"
	end
	return true
end

function _M.DispatchConfig(id, cfg_id)
	if not CheckDeviceOnline(id) then
		return false,"模块不在线"
	end
	--TODO:
	return true
end

function _M.DispatchFirmware(id, url)
	if not CheckDeviceOnline(id) then
		return false,"模块不在线"
	end

	local info = _M.devices[id]
	if info then
		local now = GetMonoTime()
		info.keep_status_util = now + 30
	end

	local proxy = GetProxy( "/", "downloader", id )
	if not proxy then
		if info then
			info.work = "upgrade-fail"
		end
		return false, "无法连接模块升级服务"
	end

	local r,done,msg = proxy.StartDownload( "firmware", url, "upgrade-firmware" )
	proxy:Destroy()
	if not r then
		if info then
			info.work = "upgrade-fail"
		end
		return false, done or "模块不能下载固件"
	elseif not done then
		if info then
			info.work = "upgrade-fail"
		end
		return false, msg or "模块下载固件失败"
	end

	if info then
		info.work = "upgrade"
	end

	return true
end

--主控程序应该周期性（建议：3秒）调用Refresh，用来更新设备状态。
--调用Refresh用来更新内部的设备状态。

local refresh_seq_no = 0

function _M.Refresh()
	if not _M.coserver_proxy then
		_dbg.Error( "DEVICE_MANAGER", "coserver_proxy error")
		return
	end

	local r, ret = _M.coserver_proxy.GetLocalEnInfo()
	if not r or type(ret) ~= "table" then
		_dbg.Error( "DEVICE_MANAGER", "Fail to get CoServer locations for getting device online status!" )
		return
	end

	local new_devices = {}
	local drop = {}
	local new = {}

	--Increment refresh sequence no.
	refresh_seq_no = refresh_seq_no + 1
	if refresh_seq_no > 1000000 then
		refresh_seq_no = 0
	end

	local changed = false
	local c, set

	for c,set in pairs(ret) do
		local item = _M.devices[ set.id ]

		if item then
			item.refresh_seq_no = refresh_seq_no
			item.status = set.status
			if( item.status == "ready" ) then
				item.status = "on"
			end

			-- _M.devices[ set.id ].address = set.address
		else --New device found, add it.
			local id = set.id
			local k = tonumber(id:match(".*%-(%d+)$"))
			if not k then
				_dbg.Error( "DEVICE_MANAGER", "Invalid local module id of", id, "so ignore it.")
			else
				changed = true

				local status = set.status
				if status == "ready" then
					status = "on"
				end
				local name
				if(k == 0) then
					name="HDMI"
				elseif (k == 1) then
					name="SDI-1"
				elseif (k == 2) then
					name="SDI-2"
				elseif (k == 3) then
					name="SDI-3"
				end

				item = {
					status = status,
					refresh_seq_no = refresh_seq_no,
					order = k + 1,
					id = id,
					type = set.options.strm_type,
					-- address = "",
					address = id .. ".coaf",
					port = set.port or 14000,
					name = name,
					with_sub = set.options.strm_sub and true or false,
					options = set.options,
					origin = "local"
				}
				local _options = 
				{
					type = item.options.strm_type,
					url =  item.options.strm_type .."://".. item.address ..":" .. item.options.strm_port .. item.options.strm_main,
					transport = "tcp",
					connect_speed = 2,
					buffer_option = "live:200:200:500:0:0"
				}

				local _sub_options = 
				{
					type = item.options.strm_type,
					url =  "",
					transport = "tcp",
					connect_speed = 2,
					buffer_option = "live:200:200:500:0:0"
				}

				--item.manual_url = item.options.strm_type .."://".. item.address ..":" .. item.options.strm_port .. item.options.strm_main
				if item.with_sub then
					_sub_options.url = item.options.strm_type .."://".. item.address ..":" .. item.options.strm_port .. item.options.strm_sub		
				end
				item.options.imgUrl = item.address .. item.options.snap_url
				item.imgUrl = item.address .. item.options.snap_url
				--item.options.buffer_option = item.buffer_option
				--streams.AddNetworkStream( item.id, item.manual_url, item.name, item.buffer_option, item.options, true, item.sub_url)
				streams.AddNetworkStream( item.id, item.name, _options, true, item.with_sub, _sub_options)
				_M.devices[ item.id ] = item
			end
		end
	end

	local someone_offline = false

	--Check each one.
	for c, set in pairs(_M.devices) do
		if set.refresh_seq_no ~= refresh_seq_no then --Not updated, that's be offline.
			set.status = "off"
			--If it suddenly be offline, it may be reset/reboot.
			set.last_sync_time = nil
			changed = true
			someone_offline = true
		else
			if set.status == "on" then
				--TODO: period sync time?
				if not set.last_sync_time then
					if _M.SyncTime(set.id) then
						set.last_sync_time = os.time()
					end
				end
			else
				someone_offline = true
			end
		end
	end

	if not _M.ui_proxy then
		--_dbg.Error( "DEVICE_MANAGER", "ui_proxy error")
		return
	end

	if changed or someone_offline then
		if someone_offline then
			_M.ui_proxy.DetectAmbaError()
		else
			_M.ui_proxy.RunningStatus()
		end
	end
end

--使用前调用Init()，确认/更新设备信息，并初始化设备状态
function _M.Init()
	_M.coserver_client = client.New( "CoManager",nil, nil, 3)
	if not _M.coserver_client then
		_dbg.Error( "DEVICE_MANAGER", "Fail to create CoServer controller client!" )
	else
		_M.coserver_proxy = _M.coserver_client:CreateProxy("/local")
		if not _M.coserver_proxy then
			_dbg.Error( "DEVICE_MANAGER", "Fail to create CoServer proxy to /" )
		end
	end

	_M.ui_client = client.New( "ui",nil, nil, 3)
	if not _M.ui_client then
		_dbg.Error( "DEVICE_MANAGER", "Fail to create ui controller client!" )
	else
		_M.ui_proxy = _M.ui_client:CreateProxy("/")
		if not _M.ui_proxy then
			_dbg.Error( "DEVICE_MANAGER", "Fail to create ui proxy to /" )
		end
	end
	--_M.Refresh()
end

function _M.SyncTime( device_id )
	local proxy = GetProxy( "/datetime", "systemctrl", device_id )
	if not proxy then
		return false
	end
	local datetime = os.date('%Y-%m-%d %H:%M:%S', os.time())
	local r, done, err = proxy.SetDatetime(datetime)
	proxy:Destroy()
	if not r or not done then
		_dbg.Error( "DEVICE", "SyncTime error:", err)
		return false
	end		
	return true
end

function _M.SyncTimeAll()
	local k,v
	for k,v in pairs(_M.synctime_dev) do
		_M.SyncTime( k )
	end
	return true
end

return _M

