local sys_config = require 'sys_config'
local mod_Decoder = require 'mod_Decoder'
local mod_GenStream = require 'mod_GeneralStream'

local coaf = require 'coaf.core'
local _dbg = require 'coaf.debug'

require 'kmCore'
local pl = require "pl.pretty"

local function compare_list(a, b)
    a = a or {}
    b = b or {}
    for k, v in pairs(a) do
        if not b[k] or b[k].url ~= v.url then
            return false
        end
    end

    for k, v in pairs(b or {}) do
        if not a[k] or a[k].url ~= v.url then
            return false
        end
    end
    return true
end

local _M = {}

local stream_meta = {}

function stream_meta:SpecialInit()
    self.fileSource = KM_STREAMING.fileMediaSource.createNew()
    if not self.fileSource then
        _dbg.Error('FILE_SOURCE', self.id, ' fail to create file source')
        self:Destroy()
        return false
    end

    self.syncSource:BindMediaSource(self.fileSource:GetMediaSource(), true)
    return true
end

function stream_meta:SpecialDestroy()
    self.fileSource:Destroy()
    _G.g_event:Unsubscribe()
    if self.event then
        _G.g_event:Unsubscribe(self.event)
    end
    self.fileSource = nil
    _dbg.Info('FILE_SOURCE', self.id, ': destroy file source')
    return true
end

function stream_meta:GetFilePath()
	local protocol, path = self.url:match( "(%w+)://(.+)$" )
	if not protocol then
		return self.url
	else
		return path
	end
end

function stream_meta:ChangeSource(file_list, options)
    if options and options.loop ~= nil then
        self.options.loop = options.loop
    end

    local is_same = compare_list(file_list, self.file_list)
    if not is_same then
        self.file_list = file_list or self.file_list

        local use_file = self:GetPlayerStatus().file_name
        if use_file and use_file ~= "" then
            local old_idx = self.file_idx
            local use_idx
            for i, v in ipairs(self.file_list) do
                if v.url == use_file then
                    use_idx = i
                    break
                end
            end

            if use_idx then
                self.file_idx = use_idx
            else
                self.file_idx = math.min(self.file_idx, math.max(1, #self.file_list))
            end
        else
            self.file_idx = math.min(self.file_idx, math.max(#self.file_list, 1))
        end
    end
end

function stream_meta:SpecialStart()
    local url = self.file_list[self.file_idx or 1] and self.file_list[self.file_idx or 1].url
    if not url then
        _dbg.Info("FILE_SOURCE", "not url so return and not start")
        return 
    end

    --NOTE:检查所有文件是否全都有错误
    local skip
    for _, v in ipairs(self.file_list) do
        if not self.error_dict[v.url] then
            skip = true
            break
        end
    end

    if not skip then
        _dbg.Info("Can not open stream, all files are invalid.")
        self:Stop(nil, true)
        return
    end

    -- NOTE：获取录像文件存储类型: 1-磁盘盘，2-文件系统下, 适配YIDUO和D300
    local storeType = self.file_list[self.file_idx or 1] and self.file_list[self.file_idx or 1].storeType
    if not storeType then
        _dbg.Error( "FileStream", "storeType is nil" )
    elseif storeType == 2 then
        _dbg.Info( "FileStream", "system file record" )
    else 
        -- NOTE:检查是否正在录像
        if not url:match("(.+)%.ts$") then
            local path, fu = url:match("(.+)%/(.+)")
            local is_ok, status = pcall(dofile, path .. "/.META/" .. fu .. ".META")
            if not is_ok or status.stopTime == "RECORDING" then
                self.error_dict[url] = true
                _dbg.Info("File is recording, prepare to stop stream.")
                self:Stop(nil, true)
                return
            end
        end
    end

    local is_ok = self.fileSource:OpenStream(url)
    if not is_ok then
        self:Stop(nil, true)
        return
    end
    _dbg.Info('FILE_SOURCE', self.id, ': started file service,  port is :', url)
end

function stream_meta:SpecialStop(group, action)
    self.fileSource:CloseStream()
    _dbg.Info('FILE_SOURCE', self.id, ': stop file source')
    return true
end

function stream_meta:GetStreamStatus(group)
    local with_act, acts = self:CheckStartActions()
    local output_start = false
    local stream_start = false
    if with_act then
        local k, v
        for k, v in pairs(acts) do
            if k == 'output' then
                output_start = true
                stream_start = true
                break
            else
                stream_start = true
            end
        end
    end

    local ret = {
        output_start = output_start,
        stream_start = stream_start
    }

    if not self.fileSource then
        ret.live = 'closed'
        ret.resolution = '?x?'
        ret.bitrate = 0
        ret.video = 'NONE'
        ret.audio = 'NONE'
    else
        local status = self.fileSource:GetStatus()
        if not status then
            ret.live = 'error'
            ret.resolution = '?x?'
            ret.bitrate = 0
            ret.video = 'NONE'
            ret.audio = 'NONE'
        else
            ret.live = status.live and status.live.status or 'error'
            if status.video then
                ret.resolution =
                    '' ..
                    (status.video.width or '0') ..
                        'x' .. (status.video.height or '0') .. '@' .. (status.video.frameRate or '0') .. 'Hz'
                ret.bitrate = status.video.bitrate or 0
                ret.video = status.video.codec or 'H264'
            else
                ret.video = 'NONE'
                ret.resolution = '?x?'
                ret.bitrate = 0
            end

            if status.audio then
                ret.audio = status.audio.codec or 'AAC'
                ret.bitrate = ret.bitrate + (status.audio.bitrate or 0)
            else
                ret.audio = 'NONE'
            end
        end
    end
    return ret
end

function stream_meta:GetMediaSource()
    --if self.syncSource then
    --      return self.syncSource:GetMediaSource()
    if self.fileSource then
        return self.fileSource:GetMediaSource()
    else
        return nil
    end
end

function stream_meta:GetPlayerStatus(group)
    local ret = {
        total_num = #(self.file_list or {}),
        current_num = math.min(self.file_idx, #(self.file_list or {}))
    }
    if not self.fileSource then
        ret.paused = 0
        ret.seek_pos = "0"
        ret.seek_req = 0
        ret.file_name = ""
        ret.file_size = ""
        ret.file_duration = ""
        ret.live_error = ""
        ret.live_status = "media_error"
        ret.live_position = ""
    else
        local player_status = self.fileSource:GetPlayerStatus()
        if not player_status then
            ret.paused = 0
            ret.seek_pos = "0"
            ret.seek_req = 0
            ret.file_name = ""
            ret.file_size = ""
            ret.file_duration = ""
            ret.live_error = ""
            ret.live_status = "media_error"
            ret.live_position = ""
        else
            ret.paused = player_status.paused
            ret.seek_pos = player_status.seek_pos
            ret.seek_req = player_status.seek_req
            ret.file_name = player_status.file_name
            ret.file_size = player_status.file_size
            ret.file_duration = player_status.file_duration
            ret.live_error = player_status.live_error
            ret.live_status = player_status.live_status
            ret.live_position = player_status.live_position
        end
    end
    ret.play = ret.live_status == "media_open" or ret.live_status == "media_play"
    return ret
end

function stream_meta:CheckLoop()
    local len = #(self.file_list or {})
    if len == 0 then
        _dbg.Info("FILE_SOURCE", "Not check loop because len = 0")
        return
    end
    self:Stop(nil, true)

    if self.file_list[self.file_idx + 1] then
        self.file_idx = self.file_idx + 1
        self:Start()
    elseif (self.file_idx + 1) > len and self.options.loop then
        self.file_idx = 1
        self:Start()
    end
end

function stream_meta:SetSeek(seek_pos)
    print("in meta, setseek", seek_pos)
    self.fileSource:FileSeek(seek_pos)
end

function stream_meta:Pause(en)
    self.fileSource:Pause(en)
end

stream_meta = mod_GenStream.ConstructStreamMeta(stream_meta)

----------------------------------

function _M.Init()
end

function _M.Create(id, buffer_option)
    local stream = mod_GenStream.CheckStreamExist(id)
    if stream then
        _dbg.Warn('FILE_SOURCE', "The creating stream '" .. id .. "' is already exist.")
        return stream
    end

    if not buffer_option or buffer_option == '' then
        buffer_option = 'live:200:200:500:0:0'
    end

    local min_group, decoder_group = mod_Decoder.AllocateDecoderGroup()
    stream = {
        id = id,
        --type = "unknown",
        type = 'file',
        decoder_group_id = min_group,
        decoderGroup = decoder_group,
        decoder_channel = nil,
        start_actions = {},
        audio_outputs = {},
        audio_enabled = {},
        vpss_using = {},
        current_audio_intfs = {},
        stream_openned = false,
        buffer_option = buffer_option,
        file_list = {},
        file_idx = 1,
        options = {},
        origin = "manual",
        error_dict = {},
        stream_options = nil
    }

    setmetatable(stream, stream_meta)
    stream:Init()

    stream.fileSource:HandleEvent("codec", id)

    stream.event = _G.g_event:Subscribe(stream.id, function(_, tbl)
        if tbl.action == "PlayerStatus" and (tbl.status == "media_end" or tbl.status == "media_exit") then
            stream.fileSource:CloseStream()
            stream:CheckLoop()
        elseif tbl.action == "PlayerStatus" and tbl.status == "media_error" then
            stream.error_dict[tbl.file] = true
        end
    end)
    return stream
end

return _M
