local sys_config = require "sys_config"
local mod_Decoder = require "mod_Decoder"
local mod_GenStream = require "mod_GeneralStream"
local audio_enc = require "mod_AudioEncoder"
local periphery = require "hiperiphery"

local coaf = require "coaf.core"
local _dbg = require "coaf.debug"

require "kmCore"

local _M = {
}

---------------------------------------------
--NOTE by左振宇：这个模块与 mod_NetStream
--大同小异。主要的差别在于 SpecialInit 和
--SpecialDestroy 的实现。这里使用了一个媒体
--切换器，用于切换音频源选择。
---------------------------------------------

local function FormatURL( url, user, pass )
	local protocol, orig_url_tail
	local user_in_url, pass_in_url

	local s

	protocol, s = url:match( "(%w+)://(.+)$" )
	if not protocol then
		return url:match("(%w+):.*"), url
	end

	protocol = protocol:lower()
	user_in_url, pass_in_url = s:match( "([^/]-):([^/]*)@.+$" )
	if not user_in_url then
		local s1 = s:match("(.-)/(.+)")
		if s1 then
				user_in_url = s1:match( "([^/]-)@./+$" )
		end
	end
	if user_in_url then
		orig_url_tail = s:match( ".-@(.+)$" )
	else
		orig_url_tail = s
	end

	if user and user ~= "" then
		user_in_url = user
		pass_in_url = pass
	end

	if user_in_url and user_in_url ~= "" then
		url = protocol .. "://" .. user_in_url
		if pass_in_url then
			url = url .. ":" .. pass_in_url
		end
		url = url .. "@" .. orig_url_tail
	else
		url = protocol .. "://" .. orig_url_tail
	end
	return protocol, url
end

local stream_meta = {}

function stream_meta:SelectAudioSource(src)
	if self.mediaSW then
		--NOTE by左振宇：请规范一下参数，现在的值太乱了。建议:
		--无音频：none
		--内嵌音频：embedded
		--模拟音频：linein
		
		if not src or src == "" or src:lower() == "none" or src:lower() == "no" then
			_dbg.Info( "AMBA_STREAM", "Select audio source '", src, "', audio is disabled." )
			self.mediaSW:SelectTrack( "audio", nil )
		elseif src:lower() == "embedded" or src:lower() == "number" then
			_dbg.Info( "AMBA_STREAM", "Select audio source '", src, "', audio is from digital embedded." )
			self.mediaSW:SelectTrack( "audio", self.networkSource:GetMediaSource() )
		elseif src:lower() == "usb" then
			_dbg.Info( "AMBA_STREAM", "Select audio source '", src, "', audio is from USB." )
			if not audio_enc or not audio_enc.usb_channel then
				_dbg.Warn( "AMBA_STREAM", "USB audio encoding channel is not initialized while selecting USB audio!" )
			else
				self.mediaSW:SelectTrack( "audio", nil ,true)
				--self.mediaSW:WaitSwitchDone(500)	--Wait 100ms
				periphery.sleep_ms(100)
				audio_enc.usb_channel:Start()
				self.mediaSW:SelectAudioEncoder( audio_enc.usb_channel:GetEncoder() )
			end
		else
			_dbg.Info( "AMBA_STREAM", "Select audio source '", src, "', audio is from local." )
			if not audio_enc or not audio_enc.channel then
				_dbg.Warn( "AMBA_STREAM", "Local audio encoding channel is not initialized while selecting local audio!" )
			else
				self.mediaSW:SelectTrack( "audio", nil, true)
				--self.mediaSW:WaitSwitchDone(1000)	--Wait 100ms
				periphery.sleep_ms(100)
				audio_enc.channel:Start()
				self.mediaSW:SelectAudioEncoder( audio_enc.channel:GetEncoder() )
			end
		end
		return true
	else
		return false
	end
end

function stream_meta:UpdateAudio(cfg)
	if not cfg then
		local mod_encoder = require "mod_Encoder"
		cfg = mod_encoder.encoders_config[self.id].audio
	end

	if (cfg) then
		self:SelectAudioSource( cfg.source )
	end
end

function stream_meta:SpecialInit()
	--Step2: Create network media source
	self.networkSource = KM_STREAMING.networkSource.createNew()
	if not self.networkSource then
		_dbg.Error( "AMBA_STREAM", self.id, ": Fail to create network media puller" )
		self:Destroy()
		return false
	end

    --Create media switcher
	self.mediaSW = KM_STREAMING.SwitchSource.createNew()

	--TODO: NOTE here:
	--  If (video) codec specified '*', that track must wait selected track be ready.
	--  And audio is specified by default 'AAC' for this application, so, if no audio in
	--  selected media source, it *FAKE* a AAC audio track (but could not get any stream data)
	self.mediaSW:SetTrackDefaultCodec( "video", "*" )
	self.mediaSW:SetTrackDefaultCodec( "audio", "MPEG4-GENERIC", {
		sampling = 48000,
		channels = 2,
		sampleSize = 1024,
		sampleWidth = 16,
		aac_format = 0, --TODO:
		aac_profile = 1, --TODO:
		aac_config = "\x11\x90"
	})

	_dbg.Info( "AMBA_STREAM", "Media switcher is created." )

	--After media switcher synchornizer (to correct-up timestamp)
	_dbg.Info( "AMBA_STREAM", "Set the media synchronizer to by-pass mode for local stream switcher" )
	self.syncSource:SetBypass()    --This is only a bypass using.

	_dbg.Info( "AMBA_STREAM", "Selected video source from Amba module" )
	self.mediaSW:SelectTrack("video", self.networkSource:GetMediaSource())

	--NOTE by左振宇：初始化时，默认选择从Amba模块来的音频。但后续应该在合适的时机依据配置参数选择正确的音频。
	self.mediaSW:SelectTrack("audio", self.networkSource:GetMediaSource())
	
	self.syncSource:BindMediaSource( self.mediaSW:GetMediaSource() )

	self:UpdateAudio()
	return true
end

function stream_meta:SpecialDestroy()
	if self.mediaSW then
		self.mediaSW:RemoveTrack("video")
		self.mediaSW:RemoveTrack("audio")
		_dbg.Info( "AMBA_STREAM", "Removed all video/audio tracks" )
	end

	--TODO: ??
	-- self.mediaSW = nil
	
	self.networkSource = nil
end

function stream_meta:CheckURLAndOptions(url, options)
	options = options or {}
	local stream_options = self.stream_options or {}
	local url_type, url = FormatURL(url, options.username, options.password) 
	if not url_type then
		_dbg.Error( 'AMBA_STREAM', self.id, "Invalid URL type of:", url )
		return false, nil
	end

	--TODO:
	stream_options.read_timeout=5000000
	stream_options.connect_timeout=5000000
	stream_options.connIntv=3000000

	local updated = false
	if self.url_type ~= url_type then
		self.url_type = url_type
		updated = true
	end

	if url_type == "rtsp" then
		local tp = options.transport or "tcp"
		if stream_options.rtsp_transport ~= tp then
			stream_options.rtsp_transport = tp
			updated = true
		end
	elseif url_type == "rtmp" then
		stream_options.rtsp_transport = nil
	elseif url_type == "udp" then
		stream_options.rtsp_transport = nil
	elseif url_type == "rtp" then
		stream_options.rtsp_transport = nil
		--TODO: Generate the SDP.
	elseif url_type == "http" then
		stream_options.rtsp_transport = nil
	else
		stream_options.rtsp_transport = nil
	end

	self.stream_options = stream_options

	return updated, url
end

function stream_meta:ChangeSource(url,options)
	local updated
	url = url or self.url or ""
	if options then
		updated,url = self:CheckURLAndOptions(url,options)
		if not url then
			return
		end
	end

	if self.url ~= url or updated then
		self.url = url
		if not self:CheckStartActions() then
			return
		end
		self:Stop(nil, true) --Stop all but don't free decoder.
		self:Start()
	end
end

function stream_meta:SpecialStart(group, action)
	self.networkSource:Open( self.url, self.stream_options )
	_dbg.Info( "NET_STREAM", self.id, ": Created stream puller for url:", self.url )
	return true
end

function stream_meta:SpecialStop(group, action)
	self.networkSource:Close()
	_dbg.Info( "NET_STREAM", self.id, ": Destroied stream puller" )
	return true
end

function stream_meta:GetStreamStatus(group)
	local with_act, acts = self:CheckStartActions()
	local output_start = false
	local stream_start = false
	if with_act then
		local k,v
		for k,v in pairs(acts) do
			if k == "output" then
				output_start = true
				stream_start = true
				break
			else
				stream_start = true
			end
		end
	end

	local ret = {
		output_start = output_start,
		stream_start = stream_start
	}

	--TODO: 通过CoAF检查 LOCAL-00/01/02/03 的视频状态和信息
	--NOTE by左振宇
	
	if not self.networkSource then
		ret.live = "closed"
		ret.resolution = "?x?"
		ret.bitrate = 0
		ret.video = "NONE"
		ret.audio = "NONE"
	else
		local status = self.networkSource:GetStatus()
		if not status then
			ret.live = "error"
			ret.resolution = "?x?"
			ret.bitrate = 0
			ret.video = "NONE"
			ret.audio = "NONE"
		else
			ret.live = status.live and status.live.status or "error"
			if status.video then
				ret.resolution = '' .. (status.video.width or "0") .. "x" .. (status.video.height or "0") .. "@" .. (status.video.frameRate or "0") .. "Hz"
				ret.bitrate = status.video.bitrate or 0
				ret.video = status.video.codec or "H264"
			else
				ret.video = "NONE"
				ret.resolution = "?x?"
				ret.bitrate = 0
			end

			if status.audio then
				ret.audio = status.audio.codec or "AAC"
				ret.bitrate = ret.bitrate + (status.audio.bitrate or 0)
			else
				ret.audio = "NONE"
			end
		end
	end
	return ret
end

function stream_meta:GetMediaSource()
	if self.syncSource then --NOTE: 返回同步（修正时间戳后）的媒体源
		return self.syncSource:GetMediaSource()
	--if self.networkSource then
	--	return self.networkSource:GetMediaSource()
	else
		return nil
	end
end

function stream_meta:GetOrigin()
	return "LOCAL INPUT"
end

function stream_meta:GetSnapUrl()
end

stream_meta = mod_GenStream.ConstructStreamMeta( stream_meta )

----------------------------------

function _M.Init()
end

function _M.Create(id, buffer_option)
	local stream = mod_GenStream.CheckStreamExist(id)
	if stream then
		_dbg.Warn( "AMBA_STREAM", "The creating stream '" .. id .. "' is already exist." )
		return stream
	end

	if not buffer_option or buffer_option == "" then
		buffer_option = "live:100:120:200:0:0"
	end

	local min_group, decoder_group = mod_Decoder.AllocateDecoderGroup()
	stream = {
		id = id,

		type = "rtsp",
		origin = "local",
		url_type = nil,
		islocal = true,
		decoder_group_id = min_group,
		decoderGroup = decoder_group,
		decoder_channel = nil,
		start_actions = {},
		audio_outputs = {},
		audio_enabled = {},
		vpss_using = {},
		current_audio_intfs = {},
		stream_openned = false,
		buffer_option = buffer_option,
		url = nil,
		stream_options = nil,
	}

	setmetatable( stream, stream_meta )
	stream:Init()
	return stream
end

return _M

