local sys_config = require "sys_config"
local mod_Decoder = require "mod_Decoder"
local mod_GenStream = require "mod_GeneralStream"
local audio_enc = require "mod_AudioEncoder"

local coaf = require "coaf.core"
local _dbg = require "coaf.debug"

require "kmCore"

local _M = {
}

---------------------------------------------
--NOTE by左振宇：这个模块与 mod_NetStream
--大同小异。主要的差别在于 SpecialInit 和
--SpecialDestroy 的实现。这里使用了一个媒体
--切换器，用于切换音频源选择。
---------------------------------------------

local stream_meta = {}

function stream_meta:SelectAudioSource(src)
	if self.mediaSW then
		--NOTE by左振宇：请规范一下参数，现在的值太乱了。建议:
		--无音频：none
		--内嵌音频：embedded
		--模拟音频：linein
		
		if not src or src == "" or src:lower() == "none" or src:lower() == "no" then
			_dbg.Info( "OUTPUT_STREAM", "Select audio source '", src, "', audio is disabled." )
			self.mediaSW:SelectTrack( "audio", nil )

		--elseif src:lower() == "embedded" or src:lower() == "number" then
		--	_dbg.Info( "OUTPUT_STREAM", "Select audio source '", src, "', audio is from digital embedded." )
		--	self.mediaSW:SelectTrack( "audio", self.networkSource:GetMediaSource() )
		else
			_dbg.Info( "OUTPUT_STREAM", "Select audio source '", src, "', audio is from mixer." )
			if self.id:match("OUTPUT2") then
				if not audio_enc or not audio_enc.vs1_channel then
					_dbg.Warn( "OUTPUT_STREAM", "Virtual screen mixer (1) audio encoding channel is not initialized while selecting mixer audio!" )
				else
					audio_enc.vs1_channel:Start()
					self.mediaSW:SelectAudioEncoder( audio_enc.vs1_channel:GetEncoder() )
				end
			else
				if not audio_enc or not audio_enc.vs0_channel then
					_dbg.Warn( "OUTPUT_STREAM", "Virtual screen mixer (0) audio encoding channel is not initialized while selecting mixer audio!" )
				else
					audio_enc.vs0_channel:Start()
					self.mediaSW:SelectAudioEncoder( audio_enc.vs0_channel:GetEncoder() )
				end
			end
		end
		return true
	else
		return false
	end
end

function stream_meta:UpdateAudio(cfg)
	if not cfg then
		local mod_encoder = require "mod_Encoder"
		cfg = mod_encoder.encoders_config[self.id].audio
	end

	if (cfg) then
		self:SelectAudioSource( cfg.source )
	else
		_dbg.Info( "OUTPUT_STREAM", "Cfg error, disable the audio source." )
		self.mediaSW:SelectTrack( "audio", nil )
	end

end

function stream_meta:SpecialInit()
	self.hiEncSource = KM_STREAMING.HISI.Encoder.createNew( self.channel - 1 )
	if not self.hiEncSource then
		_dbg.Error( "OUTPUT_STREAM", "Fail to create Hi-encoder source! ID:", self.id, "Channel:", self.channel )
		self:Destroy() 
		return false
	end

	if not self.hiEncSource:Init( false, 4*1024*1024, nil ) then
		_dbg.Error( "OUTPUT_STREAM", "Fail to initialize Hi-encoder source, ID:", self.id, "Channel:", self.channel )
		self.hiEncSource = nil
		self:Destroy()
		return false
	end

    --Create media switcher
	self.mediaSW = KM_STREAMING.SwitchSource.createNew()

	--TODO: NOTE here:
	--  If (video) codec specified '*', that track must wait selected track be ready.
	--  And audio is specified by default 'AAC' for this application, so, if no audio in
	--  selected media source, it *FAKE* a AAC audio track (but could not get any stream data)
	_dbg.Info( "OUTPUT_STREAM", "Media switcher is created." )

	self.mediaSW:SetTrackDefaultCodec( "video", "*" )
	self.mediaSW:SetTrackDefaultCodec( "audio", "MPEG4-GENERIC", {
		sampling = 48000,
		channels = 2,
		sampleSize = 1024,
		sampleWidth = 16,
		aac_format = 0, --TODO:
		aac_profile = 1, --TODO:
		aac_config = "\x11\x90"
	})

	_dbg.Info( "OUTPUT_STREAM", "Select video source from Hi-encoder for ID:", self.id, "Channel:", self.channel )
	self.mediaSW:SelectTrack("video", self.hiEncSource:GetMediaSource())

	-- 默认nil
	self.mediaSW:SelectTrack( "audio", nil )
	-- 另外一种方案，会提高耦合度
	-- self:UpdateAudio()

	--After media switcher synchornizer (to correct-up timestamp)
	_dbg.Info( "OUTPUT_STREAM", "Set the media synchronizer to by-pass mode for local stream switcher:", self.id, "Channel:", self.channel )
	self.syncSource:SetBypass()    --This is only a bypass using.

	self.syncSource:BindMediaSource( self.mediaSW:GetMediaSource() )
	return true
end

function stream_meta:SpecialDestroy()
	if self.mediaSW then
		self.mediaSW:RemoveTrack("video")
		self.mediaSW:RemoveTrack("audio")
		_dbg.Info( "OUTPUT_STREAM", "Removed all video/audio tracks for:", self.id )
	end

	if self.hiEncSource then
		self.hiEncSource:Stop()
		self.hiEncSource = nil
	end
end

function stream_meta:SpecialStart(group, action)
	self.hiEncSource:Start()
	_dbg.Info( "OUTPUT_STREAM", self.id, ": Created stream puller!" )
	return true
end

function stream_meta:SpecialStop(group, action)
	--self.hiEncSource:Stop()
	_dbg.Info( "OUTPUT_STREAM", self.id, ": Destroied stream puller" )
	return true
end

function stream_meta:GetStreamStatus(group)
	local with_act, acts = self:CheckStartActions()
	local output_start = false
	local stream_start = false
	if with_act then
		local k,v
		for k,v in pairs(acts) do
			if k == "output" then
				output_start = true
				stream_start = true
				break
			else
				stream_start = true
			end
		end
	end

	local ret = {
		output_start = output_start,
		stream_start = stream_start
	}

	if not self.hiEncSource then
		ret.live = "closed"
		ret.resolution = "?x?"
		ret.bitrate = 0
		ret.video = "NONE"
		ret.audio = "NONE"
	else
	--TODO: 更新！视频状态和信息
		ret.live = "connected"
		ret.resolution = "1920x1080@60Hz"
		ret.bitrate = 4000

		local mo_encoder = require "mod_Encoder"
		local encoderobj = mo_encoder.local_out_encoders[ self.id ]

		if( encoderobj and encoderobj.encoder ) then
			ret.resolution = '' .. encoderobj.encoder.encodingOpt.attr.u32PicWidth .. "x" .. encoderobj.encoder.encodingOpt.attr.u32PicHeight
			ret.resolution = ret.resolution .. "@" .. encoderobj.encoder.bitrateOpts["CBR"].attr.fr32DstFrmRate .. "Hz"
			--TODO: Get actual bitrate
			ret.bitrate = encoderobj.encoder.bitrateOpts[ encoderobj.encoder.currentBrMode or "CBR" ].attr.u32BitRate
		end

		--TODO: Get audio selection
		ret.audio = "AAC"
		ret.video = "H264"
	end
	return ret
end

function stream_meta:GetMediaSource()
	if self.syncSource then --NOTE: 返回同步（修正时间戳后）的媒体源
		return self.syncSource:GetMediaSource()
	--if self.hiEncSource then
	--	return self.hiEncSource:GetMediaSource()
	else
		return nil
	end
end

stream_meta = mod_GenStream.ConstructStreamMeta( stream_meta )

----------------------------------

function _M.Init()
end

function _M.Create(id, channel, name)
	local stream = mod_GenStream.CheckStreamExist(id)
	if stream then
		_dbg.Warn( "OUTPUT_STREAM", "The creating stream '" .. id .. "' is already exist." )
		return stream
	end

	--Not decodable.
	--local min_group, decoder_group = mod_Decoder.AllocateDecoderGroup()
	stream = {
		id = id,
		name = name,
		channel = channel,
		origin = "output",
		type = "output",
		islocal = true,
		url = id, --Not actual use.
		--decoder_group_id = min_group,
		--decoderGroup = decoder_group,
		--decoder_channel = nil,
		buffer_option = "live:50:100:200:0:0",
		start_actions = {},
		audio_outputs = {},
		audio_enabled = {},
		vpss_using = {},
		current_audio_intfs = {},
		stream_openned = false,
		stream_options = nil,
	}

	setmetatable( stream, stream_meta )
	stream:Init()
	return stream
end

return _M

