local coaf = require "coaf.core"
local _dbg = require "coaf.debug"

local periphery = require "hiperiphery"
local I2C = periphery.I2C
local HIMM = periphery.HIMM

local _M = {
	i2c_dev = "scl=0x121c0000:4 sda=0x121c0000:5",

	PINMUX_CTRL = {
		scl = { muxreg=0x120f0140, pingroup=0x121c0000, pin=4, muxval=0 },	--GPIO7_4: SCL
		sda = { muxreg=0x120f0144, pingroup=0x121c0000, pin=5, muxval=0 },	--GPIO7_5: SDA
	},

	BASE_ADDR = 0xe2,
	__pin_init = false
}

function _M.InitPinmux()
	if _M.__pin_init then
		return
	end

	local pinmux_ctrl = _M.PINMUX_CTRL
	local seqs = { "scl", "sda" }
	local i
	local val
	for i=1, #seqs do
		if type(pinmux_ctrl[ seqs[i] ]) == "table" then
		--Only if configured.
			--NOTE: Sometimes, the PIN has no delcared mux control register.
			if pinmux_ctrl[ seqs[i] ].muxreg and pinmux_ctrl[ seqs[i] ].muxreg ~= 0 then
				HIMM.write( pinmux_ctrl[ seqs[i] ].muxreg, pinmux_ctrl[ seqs[i] ].muxval )
			end

			val = HIMM.read( pinmux_ctrl[ seqs[i] ].pingroup + 0x400 )
			--Make default as input.
			val = val & ~(1<<pinmux_ctrl[ seqs[i] ].pin)
			HIMM.write( pinmux_ctrl[ seqs[i] ].pingroup + 0x400, val )
		--End.
		else
			pinmux_ctrl[ seqs[i] ] = nil
		end
	end
	_M.__pin_init = true
end

function _M.Init()
	_M.InitPinmux()

	local done,err = pcall( function() _M.i2c_handle = I2C( _M.i2c_dev ) end )
	if not done then
		if type(err) == "table" then
			_dbg.Error( "FPGA_CONTROL", "Could not open I2C device for FPGA controlling. ErrMsg:", err.message or "---" )
		else
			_dbg.Error( "FPGA_CONTROL", "Could not open I2C device for FPGA controlling. ErrMsg:", err )
		end
	else
		_dbg.Info( "FPGA_CONTROL", "FPGA controller openned, at " .. string.format("%02x",_M.BASE_ADDR) .. "@" .. _M.i2c_dev )
	end
end

function _M.ReadReg( addr, itemArr )
	if not _M.i2c_handle then
		return nil, "Invalid I2C handle"
	end
	if not itemArr or type(itemArr) ~= "table" then
		itemArr = { 0xff, flags = I2C.I2C_M_RD }
	else
		if #itemArr < 1 then
			itemArr[1] = 0xff
		end
		if not itemArr.flags then
			itemArr.flags = I2C.I2C_M_RD
		else
			itemArr.flags = itemArr.flags | I2C.I2C_M_RD
		end
	end

	local msg = { {addr}, itemArr }
	local done,err = pcall( function() _M.i2c_handle:transfer( _M.BASE_ADDR>>1, msg ) end )
	if not done then
		if type(err) == "table" then
			--print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err.message or "---" )
			return nil, err.message or "---"
		else
			--print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err )
			return nil, err
		end
	else
		return itemArr
	end
end

-----------------------------------

function _M.WriteReg( addr, itemArr )
	if not _M.i2c_handle then
		return false, "Invalid I2C handle"
	end
	if not itemArr or type(itemArr) ~= "table" or #itemArr < 1 then
		return false, "Invalid data"
	end

	table.insert( itemArr, 1, addr )
	local msg = { itemArr }
	local done,err = pcall( function() _M.i2c_handle:transfer( _M.BASE_ADDR>>1, msg ) end )
	if not done then
		if type(err) == "table" then
			--print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err.message or "---" )
			return nil, err.message or "---"
		else
			--print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err )
			return nil, err
		end
	else
		return true
	end
end

return _M

