require "himpp"
local _dbg = require "coaf.debug"

local SYS = HiMPP.SYS
local AO = HiMPP.AO
local ACODEC = HiMPP.ACODEC

local _M = {
	MAX_CHANNELS = 3,
	aDev = "/dev/acodec",
	aCodec = nil,
	channels = {}
}

function _M.Init(resetCodec)
	if _M.aCodec then
		_dbg.Warn( "AOUT", "The audio out module is already initialized.")
		return
	end

	_M.aCodec = ACODEC.CreateInnerCodec( _M.aDev )
	if not _M.aCodec then
		_dbg.Error( "AOUT", "Fail to create audio Codec!" )
		return
	end

	if _M.aCodec:Open() < 0 then
		_dbg.Error( "AOUT", "Fail to open audio Codec!" )
		_M.aCodec = nil
		return
	end

	if resetCodec then
		_M.aCodec:Reset()
	end

	_M.aCodec:Close()
end

function _M.Cleanup()
	local k, chanObj
	for k, chanObj in pairs(_M.channels) do
		if chanObj then
			chanObj:Destroy()
		end
	end
	_M.channels = {}

	if _M.aCodec then
		_M.aCodec:Open()
		_M.aCodec:Reset()
		_M.aCodec:Close()
		_M.aCodec = nil
	end
end

--------------------------------------------------

local channel_meta = {}

-- Options:
-- {
-- 		channel = "MONO"/"STEREO",
-- 		sampleRate = ...,
-- 		sampleWidth = 8/16/24,
-- 		sampleSize = 256/512/...
-- }

function channel_meta:SetAudioOption( opts )
	local changed = false
	local srChanged = false
	if opts.channel and opts.channel ~= self.aoPubAttr.enSoundmode then
		self.aoPubAttr.enSoundmode = opts.channel
		changed = true
	end

	if opts.sampleRate and opts.sampleRate ~= self.aoPubAttr.enSamplerate then
		self.aoPubAttr.enSamplerate = opts.sampleRate
		changed = true
		srChanged = true
	end

	if opts.sampleWidth and opts.sampleWidth ~= self.aoPubAttr.enBitwidth then
		self.aoPubAttr.enBitwidth = opts.sampleWidth
		changed = true
	end

	if opts.sampleSize and opts.sampleSize ~= self.aoPubAttr.u32PtNumPerFrm then
		self.aoPubAttr.u32PtNumPerFrm = opts.sampleSize
		changed = true
	end

	if srChanged and self.connectedDAC then
		_M.aCodec:Open()
		_M.aCodec:SetSampleRate( self.aoPubAttr.enSamplerate )
		_M.aCodec:Close()
	end

	if changed and self.initialized then
		if self.started then
			self.aoChannel:Disable()
		end
		self.aoDev:Disable()
		local r, err = self.aoDev:SetPubAttr( self.aoPubAttr )
		if not r then
			_dbg.Error( "AOUT", "Fail to configure audio OUT device, error:", err )
		end

		self.aoDev:Enable()
		if self.started then
			self.aoChannel:Enable()
		end

	end
end

function channel_meta:Init()
	if self.initialized then
		return
	end

	--Only related to inner DAC should set this.
	if self.connectedDAC then
		_M.aCodec:Open()
		if _M.aCodec:SetSampleRate( self.aoPubAttr.enSamplerate ) ~= 0 then
			_dbg.Warn( "AOUT", "Fail to set audio Codec chip sample rate to ", self.aoPubAttr.enSamplerate )
		end
		_M.aCodec:Close()
	end

	local r, err = self.aoDev:SetPubAttr( self.aoPubAttr )
	if not r then
		_dbg.Error( "AOUT", "Fail to configure audio OUT device, error:", err )
		return
	end

	r = self.aoDev:Enable()
	if r ~= 0 then
		_dbg.Error( "AOUT", "Fail to enable audio OUT device! error code:", string.format("%08X",r) )
		return
	end

	self.aoChannel = self.aoDev:GetChannel(0)
	self.initialized = true
end

function channel_meta:Destroy()
	if self.initialized then
		self.aoChannel:Disable()
		self.aoChannel = nil
		self.aoDev:Disable()
		self.initialized = false
	end
end

function channel_meta:Start()
	if not self.initialized then
		self:Init()
	end

	if self.initialized then
		self.aoChannel:Enable()
		self.started = true
	end
end

function channel_meta:Stop()
	if self.initialized then
		self.aoChannel:Disable()
		self.started = false
	end
end

function channel_meta:SetGain( gain )
	if not self.initialized then
		self:Init()
		if not self.initialized then
			return
		end
	end
	self.aoDev:SetVolume(gain)
end

function channel_meta:SetAnalogGain( gain )
	--Only connected DAC support this.
	if self.connectedDAC then
		_M.aCodec:Open()
		_M.aCodec:SetDACGain( gain, gain )
		_M.aCodec:Close()
	end
end

function channel_meta:Mute()
	--if self.connectedDAC then
	--	_M.aCodec:Open()
	--	_M.aCodec:SetDACMute( true, true )
	--	_M.aCodec:Close()
	--else
		if not self.initialized then
			self:Init()
			if not self.initialized then
				return
			end
		end
		self.aoDev:SetMute(true)
	--end
end

function channel_meta:Unmute()
--	if self.connectedDAC then
--		_M.aCodec:Open()
--		_M.aCodec:SetDACMute( false, false )
--		_M.aCodec:Close()
--	else
		if not self.initialized then
			self:Init()
			if not self.initialized then
				return
			end
		end
		self.aoDev:SetMute(false)
--	end
end

channel_meta.__index = channel_meta

---------------------------------------------------

function _M.CreateChannel(id, connDAC)
	if id < 1 or id > _M.MAX_CHANNELS then
		_dbg.Error( "AOUT", "Create channel", id, "is out of supported channel range [ 1,", _M.MAX_CHANNELS, "]!" )
		return nil
	end

	if _M.channels[id] then
		return _M.channels[id]
	end

	local chanObj = {
		id = id,
		connectedDAC = connDAC,
		aoDev = AO.GetDevice(id-1),
		aoChannel = nil,
		started = false,

		aoPubAttr = {
			enSamplerate = 48000,
			enBitwidth = 16,
			enWorkmode = "I2S_MASTER",
			enSoundmode = "STEREO",
			u32EXFlag = 0,
			u32FrmNum = 6, --Don't be too large.
			u32PtNumPerFrm = 1024,
			u32ClkChnCnt = 2,
			u32ChnCnt = 2,
			u32ClkSel = 0
		}
	}

	local r = setmetatable( chanObj, channel_meta )
	_M.channels[id] = r

	return r
end

---------------------------------------------------

return _M


