require "himpp"
local _dbg = require "coaf.debug"

local SYS = HiMPP.SYS
local VPSS = HiMPP.VPSS
local VDEC = HiMPP.VDEC
local VB = HiMPP.VB

local MAX_RES_1080P = nil
local MAX_VIDEO_WIDTH = 3840
local MAX_VIDEO_HEIGHT = 2176
local MAX_VB_BLK_SIZE = 3840*3264 --Aka. 3840*2160*3/2
local MAX_VB_BLK_COUNT = 32
local MAX_PMV_BLK_SIZE = 3840*540
local MAX_PMV_BLK_COUNT = 24

local _x, features = pcall(require, "features")
if _x and type(features) == "table" then
	if features.FEATURE_MAX_RES_1080P then
		MAX_RES_1080P = true
		MAX_VIDEO_WIDTH = 1920
		MAX_VIDEO_HEIGHT = 1088
		MAX_VB_BLK_SIZE = 1920*1088*2
		MAX_VB_BLK_COUNT = 64
		MAX_PMV_BLK_SIZE = 120*68*64*2
		MAX_PMV_BLK_COUNT = 32
	end
else
	features = nil
end

--NOTE 1: Create decoder channel, specify ddr name.

local _M = {
	-- You can change this for VPSS alocation.
	VPSS_ID_BASE = 2,
	MAX_CHANNELS = 16,
	channels = {}
}

function _M.Init()
	local i
	-- Force free channels, for new using.
	for i=1, _M.MAX_CHANNELS do
		VDEC.ForceFreeChannel( i-1 )
	end

	VB.SetModPoolConf( "VDEC", 2, {
		{ u32BlkSize = MAX_VB_BLK_SIZE, u32BlkCnt = MAX_VB_BLK_COUNT },
		{ u32BlkSize = MAX_PMV_BLK_SIZE, u32BlkCnt = MAX_PMV_BLK_COUNT },
	})

	VB.InitModCommPool("VDEC")
end

function _M.Cleanup()
	local i
	for i=1, _M.MAX_CHANNELS do
		if _M.channels[i] then
			_M.channels[i]:Destroy()
		end
	end
	_M.channels = {}
end

-----------------------------------------------------------
-- Decoder meta
-----------------------------------------------------------

local channel_meta = {}

function channel_meta:Init()
	if self.initialized then
		return
	end

	self.vdecChannel = VDEC.GetChannel( self.id - 1) 
	if not self.vdecChannel then
		_dbg.Error( "DECODER", "Fail to create decoder channel for ID=" .. self.id )
		return
	end

	self.vdecChannel:SetVBCnt( self.decoderAttr.attr.u32RefFrameNum + 4 ) -- NOTE: here.
	SYS.SetMemConf( self.vdecChannel, self.ddrSpec )

	local r, errmsg = self.vdecChannel:Create(self.decoderAttr)
	if not r then
		_dbg.Error( "DECODER", "Fail to initialize decoder channel for ID=" .. self.id .. ", error:", errmsg )
		self.vdecChannel = nil
		return
	end

	self.vdecChannel:SetDisplayMode( self.displayMode )
	-- For compatible reason.
	--NOTE: HI3535 not support h265 decode
	if self.codecType == "H265" then
		self.vdecChannel:SetProtocolParam({
			type = "H265",
			param = {
			s32MaxSliceSegmentNum = 136,
			s32MaxVpsNum = 16,
			s32MaxSpsNum = 16, --MAX 16
			s32MaxPpsNum = 64  --MAX 64
			}
		})
	else
		self.vdecChannel:SetProtocolParam({ 
			type = "H264",
			param = {
			s32MaxSliceNum = 136,
			s32MaxSpsNum = 32,
			s32MaxPpsNum = 64, --MAX 256
			s32DisplayFrameNum = 2
			}
		})
	end

	self.vdecChannel:SetParam( self.decoderParam )

	VPSS.ForceFreeGroup( _M.VPSS_ID_BASE + self.id - 1 )

	-- Create a VPSS goup, for resizer, also for Overlay.
	self.vpssGroup = VPSS.CreateGroup(_M.VPSS_ID_BASE + self.id - 1, {
		u32MaxW = MAX_VIDEO_WIDTH,
		u32MaxH = MAX_VIDEO_HEIGHT,
		enPixFmt = "YUV_SEMIPLANAR_420",
		--bNrEn = true,
		enDieMode = "NODIE"
	})

	if not self.vpssGroup then
		_dbg.Error( "DECODER", "Fail to create VPSS group, for video decoding channel ID=" .. self.id )
		self.vdecChannel:Destroy()
		self.vdecChannel = nil
		return
	end

	if self.doubleVpss then
		self.vpssGroup2 = VPSS.CreateGroup( _M.VPSS_ID_BASE + _M.MAX_CHANNELS + self.id - 1, {
			u32MaxW = MAX_VIDEO_WIDTH,
			u32MaxH = MAX_VIDEO_HEIGHT,
			enPixFmt = "YUV_SEMIPLANAR_420",
			--bNrEn = true,
			enDieMode = "NODIE"
		})
		if not self.vpssGroup2 then
			_dbg.Warn( "DECODER", "Fail to create second VPSS group, for video decoding channel ID=" .. self.id )
		end
	end

	local r, errmsg
	-- 2nd, Bind VDEC channel to VPSS group.
	r, errmsg = SYS.Bind( self.vdecChannel, self.vpssGroup )
	if not r then
		_dbg.Error( "DECODER", "Fail to bind decoder channel to VPSS group, for decoding channel ID=" .. self.id .. ", reason:", errmsg )
		self.vpssGroup:Destroy()
		self.vpssGroup = nil
		return
	end

	if self.vpssGroup2 then
		r, errmsg = SYS.Bind( self.vdecChannel, self.vpssGroup2 )
		if not r then
			_dbg.Warn( "DECODER", "Fail to bind decoder channel to VPSS group 2, for decoding channel ID=" .. self.id .. ", reason:", errmsg )
		end
	end

	self.initialized = true
end

function channel_meta:Destroy()
	if self.initialized then
		_dbg.Print("DECODER", "VpssGroup stop:",
		self.vpssGroup:Stop()
		)

		if self.vpssGroup2 then
			self.vpssGroup2:Stop()
		end
		
		_dbg.Print("DECODER", "vdecChannel stop:",
		self.vdecChannel:Stop()
		)

		_dbg.Print("DECODER", "Unbind:",
		SYS.UnBind( self.vdecChannel, self.vpssGroup )
		)

		_dbg.Print("DECODER", "VpssGroup destroy:",
		self.vpssGroup:Destroy()
		)

		self.vpssGroup = nil
		if self.vpssGroup2 then
			self.vpssGroup2:Destroy()
			self.vpssGroup2 = nil
		end

		_dbg.Print("DECODER", "vdecChannel destroy:",
		self.vdecChannel:Destroy()
		)

		self.vdecChannel = nil
		self.initialized = false
	end
end

function channel_meta:Start()
	if not self.initialized then
		self:Init()
	end
	if self.initialized then
		self.vdecChannel:Start()

		self.vpssGroup:Start()
		if self.vpssGroup2 then
			self.vpssGroup2:Start()
		end
	end
end

function channel_meta:Stop()
	if self.initialized then
		self.vpssGroup:Stop()
		if self.vpssGroup2 then
			self.vpssGroup2:Stop()
		end
		self.vdecChannel:Stop()
	end
end

function channel_meta:GetVPSSGroup(id)
	if not id or id < 2 then
		return self.vpssGroup
	else
		return self.vpssGroup2
	end
end

function channel_meta:GetDecChannel()
	return self.vdecChannel
end

-- attrs:
-- 		codec : Video codec (H264/H265)
-- 		width : Video width
-- 		height : Video height
-- 		framerate : Frame rate (just for reference, no used)
--		video_delay : 0 or > 0
--		ref_frames : reference frames.
function channel_meta:Configure(attrs)
	if not attrs then
		_dbg.Error( "DECODER", "Configure: Invalid decoder channel attributes" )
		return
	end
	--local pt = require "pl.pretty"
	--pt.dump(attrs)

	local changed = false
	if attrs.codec and attrs.codec ~= self.codecType then
		self.codecType = string.upper(attrs.codec)
		self.decoderAttr.enType = self.codecType
		changed = true
	end

	--注意此处：后期完善音视频编码参数动态调整情况
	self.decoderAttr.u32BufSize = MAX_VB_BLK_SIZE
	self.decoderAttr.u32PicWidth = MAX_VIDEO_WIDTH
	self.decoderAttr.u32PicHeight = MAX_VIDEO_HEIGHT
--[[
	local w = attrs.width or self.decoderAttr.u32PicWidth
	local h = attrs.height or self.decoderAttr.u32PicHeight
	if w <= 128 or h <= 128 then --Why too small?
		w = 1920
		h = 1088
	else
		w = (w+15) & 0xfff0
		h = (h+15) & 0xfff0
	end

	if w ~= self.decoderAttr.u32PicWidth or h ~= self.decoderAttr.u32PicHeight then
		changed = true
		self.decoderAttr.u32PicWidth = w
		self.decoderAttr.u32PicHeight = h
		self.decoderAttr.u32BufSize = w*h*3/2
	end
]]
	local supportB = self.supportB
	if attrs.video_delay then
		supportB = tonumber(attrs.video_delay) > 0
	end

	-- codecType or supportB changed
	if supportB ~= self.supportB or changed then
		self.supportB = supportB 
		changed = true

		if self.supportB then
			self.decoderAttr.attr.enMode = "STREAM"
			self.decoderAttr.attr.bTemporalMvpEnable = true
			self.decoderAttr.attr.s32SupportBFrame = 1
			self.decoderParam.s32DecMode = 0
			self.decoderParam.s32DecOrderOutput = 0
		else
			if self.codecType == "H264" then
				self.decoderAttr.attr.enMode = "FRAME"
				self.decoderAttr.attr.bTemporalMvpEnable = false
			else
				self.decoderAttr.attr.enMode = "STREAM"
				self.decoderAttr.attr.bTemporalMvpEnable = true
			end
			self.decoderParam.s32DecMode = 1
			self.decoderParam.s32DecOrderOutput = 0
			self.decoderAttr.attr.s32SupportBFrame = 0
		end
	end

	local rf = self.decoderAttr.attr.u32RefFrameNum
	if attrs.ref_frames then
		rf = attrs.ref_frames
		if rf < 2 then
			rf = 2
		end
	end

	if rf ~= self.decoderAttr.attr.u32RefFrameNum then
		changed = true
		self.decoderAttr.attr.u32RefFrameNum = rf
	end

	if changed and self.vdecChannel then
		--self.vdecChannel:SetAttr( self.decoderAttr )
		--self.vdecChannel:SetParam( self.decoderParam )
		_dbg.Info( "DECODER", "Configure: decoder channel attributes" )
		return true
	end
end

function channel_meta:SetDisplayMode(dm)
	if dm ~= self.displayMode then
		self.displayMode = dm
		if self.vdecChannel then
			self.vdecChannel:SetDisplayMode( dm )
		end
	end
end

channel_meta.__index = channel_meta

-----------------------------------------------------------
function _M.CreateChannel( id, codecType, doubleVpss, ddrSpec )
	id = id or 1
	if id < 1 or id > _M.MAX_CHANNELS then
		_dbg.Error( "DECODER", "Invalid decoder channel ID passed. It's only support 1-" .. _M.MAX_CHANNELS .. " channels." )
		return nil
	end

	if _M.channels[id] then
		_dbg.Info( "DECODER", "The requesting channel (ID=" .. id .. ") is already created, returns the exist channel." )
		return _M.channels[id]
	end

	local chanObj = {
		id = id,
		codecType = codecType or "H264",
		doubleVpss = doubleVpss,

		vpssGroup = nil,
		vpssGroup2 = nil,
		vdecChannel = nil,
		ddrSpec = ddrSpec,

		displayMode = "PREVIEW",
		supportB =true,

		decoderAttr = {
			enType = codecType or "H264",
			u32BufSize = MAX_VB_BLK_SIZE,
			u32Priority = 5,
			u32PicWidth = MAX_VIDEO_WIDTH,
			u32PicHeight = MAX_VIDEO_HEIGHT,
			attr = {
				enMode = "STREAM",
				u32RefFrameNum = 4,--Need sps para ->ref_frame_count
				bTemporalMvpEnable = true,
				s32SupportBFrame = 1
			}
		},

		decoderParam = {
			s32DisplayFrameNum = 2, --? 3
			s32ChanErrThr = 30,
			s32ChanStrmOFThr = 0,
			s32DecMode = 0, -- 1: IP, 0: IPB
			s32DecOrderOutput = 0, -- 0: reference pts output, 1:reference dts output
			enVideoFormat = "TILE",
			enCompressMode = "NONE" --"SEG128?"
		}
	}

	local r = setmetatable( chanObj, channel_meta )
	_M.channels[id] = r
	return r
end

function _M.DestroyChannel( id )
	if id then
		_M.channels[id] = nil
	end
end
----------------------------------

return _M

