require "himpp"
local _dbg = require "coaf.debug"

local SYS = HiMPP.SYS
local VPSS = HiMPP.VPSS
local VENC = HiMPP.VENC

local _M = {
	MAX_CHANNELS = 6,
	channels = {}
}

--NOTE: SYS.SetMemConf(...) to configure channel using which DDR.
-- Init Encoder module.
function _M.Init()
end

-- Cleanup encoder module.
function _M.Cleanup()
	for i=1, _M.MAX_CHANNELS do
		if _M.channels[i] then
			_M.channels[i]:Destroy()
		end
	end
	_M.channels = {}
end

-------------------------------------------------------
-- Channel meta methods.
-------------------------------------------------------
local channel_meta = {}

-- Init encoding channel.
function channel_meta:Init()
	if self.initialized then
		return
	end

	if not self.srcChannel then
		_dbg.Error( "ENCODER", "No valid video source channel bound with encoder " .. self.id )
		return
	end

	-- Create a video encoder channel.
	local attr = {
		encoding = self.encodingOpt,
		bitrate = self.bitrateOpts[ self.currentBrMode ]
	}

	--local pt = require "pl.pretty"
	--pt.dump(attr)

	local venc_chan = VENC.GetChannel( self.id - 1) 
	if not venc_chan then
		_dbg.Error( "ENCODER", "Fail to get VENC channel of " .. self.id )
		return nil
	end

	local r, errmsg = venc_chan:Create(attr)
	if not r then
		_dbg.Error( "ENCODER", "Fail to create VENC channel of " .. self.id .. ", with error:", errmsg )
		venc_chan = nil
		return nil
	end

	r, errmsg = SYS.Bind( self.srcChannel, venc_chan )
	if not r then
		_dbg.Error( "ENCODER", "Fail to bind video source channel to encoder, for encoding channel " .. self.id .. ", reason:", errmsg )
		venc_chan:Destroy()
		return
	end

	self.vencChannel = venc_chan
	self.initialized = true
end

-- Destroy channel.
function channel_meta:Destroy()
	if not self.initialized then
		return
	end

	self.vencChannel:Stop()
	self.srcChannel:Disable()

	SYS.UnBind( self.srcChannel, self.vencChannel )

	self.vencChannel:Destroy()
	self.vencChannel = nil
	
	self.initialized = false
end

-- Start channel encoding.
function channel_meta:Start()
	if not self.initialized then
		self:Init()
	end
	if self.initialized then
		self.srcChannel:Enable()
		self.vencChannel:Start()
		self.started = true
	end
end

-- Stop channel encoding.
function channel_meta:Stop()
	if self.initialized then
		self.vencChannel:Stop()
		self.srcChannel:Disable()
		self.started = nil
	end
end

-- Update encoding parameters.
function channel_meta:UpdateEncoding(onfly)
	if self.vencChannel then

		if self.started and not onfly then
			self.vencChannel:Stop()
		end

		local attr = {
			encoding = self.encodingOpt,
			bitrate = self.bitrateOpts[ self.currentBrMode ]
		}
		local r, errmsg = self.vencChannel:SetAttr( attr )
		if not r then
			_dbg.Error( "ENCODER", "Fail to set encoding parameters, for channel " .. self.id .. ", reason:", errmsg )
		end

		if self.started and not onfly then
			self.vencChannel:Start()
		end
	end
end

-- Set source (Aka. captured from VI) image (width,height,framerate)
-- each one can be nil so ignore that.
function channel_meta:SetSourceImage( w, h, fr )
	local changed = self._FORCE_UPDATE

	if w and w ~= self.srcImg.width then
		self.srcImg.width = w
		changed = true
	end

	if h and h ~= self.srcImg.height then
		self.srcImg.height = h
		changed = true
	end

	if fr and fr ~= self.srcImg.frameRate then
		self.srcImg.frameRate = fr
		changed = true
	end

	if not changed then return end

	self.bitrateOpts["CBR"].attr.u32SrcFrmRate = self.srcImg.frameRate
	self.bitrateOpts["VBR"].attr.u32SrcFrmRate = self.srcImg.frameRate

	if self.dstImg.frameRate < 1 then
		if self.dstImg.frameRate <= 0 or self.srcImg.frameRate < 50 then
			self.bitrateOpts["CBR"].attr.fr32DstFrmRate = self.srcImg.frameRate
			self.bitrateOpts["VBR"].attr.fr32DstFrmRate = self.srcImg.frameRate
		else
			self.bitrateOpts["CBR"].attr.fr32DstFrmRate = self.srcImg.frameRate * self.dstImg.frameRate
			self.bitrateOpts["VBR"].attr.fr32DstFrmRate = self.srcImg.frameRate * self.dstImg.frameRate
		end
	end

	--[[
	if self.vencChannel then
		self:UpdateEncoding()
	end
	]]
	self._FORCE_UPDATE = true
	self:SetEncodingImage() --All is nil, so just update.
	self._FORCE_UPDATE = false
end

-- Set encoding image (width, height, framerate) and each one
-- can be nil so ignore it.
-- If each one value is 0, means it always keep same value with
-- source image.
-- If frame rate is >0 and <1, so the actual frame rate is
-- source_framerate * value
function channel_meta:SetEncodingImage( w, h, fr )
	--local changed = self._FORCE_UPDATE
	local changed = true

	if w and w ~= self.dstImg.width then
		self.dstImg.width = w
		changed = true
	end

	if h and h ~= self.dstImg.height then
		self.dstImg.height = h
		changed = true
	end

	if fr and fr ~= self.dstImg.frameRate then
		self.dstImg.frameRate = fr
		changed = true
	end

	if not changed then return end

	local act_w, act_h
	act_w = self.dstImg.width < 16 and self.srcImg.width or self.dstImg.width
	act_h = self.dstImg.height < 16 and self.srcImg.height or self.dstImg.height

	self.encodingOpt.attr.u32PicWidth = act_w
	self.encodingOpt.attr.u32PicHeight = act_h
	
	self.bitrateOpts["CBR"].attr.u32SrcFrmRate = self.srcImg.frameRate
	self.bitrateOpts["VBR"].attr.u32SrcFrmRate = self.srcImg.frameRate

	if self.dstImg.frameRate < 1 then
		if self.dstImg.frameRate <= 0 or self.srcImg.frameRate < 50 then
			self.bitrateOpts["CBR"].attr.fr32DstFrmRate = self.srcImg.frameRate
			self.bitrateOpts["VBR"].attr.fr32DstFrmRate = self.srcImg.frameRate
		else
			self.bitrateOpts["CBR"].attr.fr32DstFrmRate = self.srcImg.frameRate * self.dstImg.frameRate
			self.bitrateOpts["VBR"].attr.fr32DstFrmRate = self.srcImg.frameRate * self.dstImg.frameRate
		end
	else
		self.bitrateOpts["CBR"].attr.fr32DstFrmRate = self.dstImg.frameRate
		self.bitrateOpts["VBR"].attr.fr32DstFrmRate = self.dstImg.frameRate
	end

	if self.vencChannel then
		self:UpdateEncoding()
	end
end

-- Set encoding profile, can be "base"/"baseline", "main", "high"
function channel_meta:SetProfile( pf )
	local changed = self._FORCE_UPDATE
	if pf then
		local oldProfile = self.encodingOpt.attr.u32Profile
		local pf_name = pf:lower()
		if pf_name == "base" or pf_name == "baseline" then
			self.encodingOpt.attr.u32Profile = 0
		elseif pf_name == "main" then
			self.encodingOpt.attr.u32Profile = 1
		else
			self.encodingOpt.attr.u32Profile = 2
		end
		if oldProfile ~= self.encodingOpt.attr.u32Profile then
			changed = true
		end
	end

	if changed and self.vencChannel then
		self:UpdateEncoding()
	end
end

-- Set encoding parameters (in a table). These parameters should be:
--		u32MaxPicWidth = 1920,
--		u32MaxPicHeight = 1088,
--		u32BufSize = 1920*1088*2,
--		u32Profile = 2, --Profile: 0: Baseline, 1: MP, 2: HP
--		u32PicWidth = (chanId == 1 and 1920 or 720),
--		u32PicHeight = (chanId == 1 and 1080 or 480),
--		u32BFrameNum = 0,
--		bByFrame = true

function channel_meta:SetEncodingParam( param )
	local changed =self._FORCE_UPDATE
	if type(param) == "table" then
		local k,v
		for k,v in pairs(param) do
			if v ~= self.encodingOpt.attr[ k ] then
				changed = true
				self.encodingOpt.attr[k] = v
			end
		end
	end
	if changed and self.vencChannel then
		self:UpdateEncoding()
	end
end

-- Set encoding bitrate options (in a table), these options should be:
-- For CBR:
--		u32Gop = 60,
--#		u32StatTime = 1,
--#		u32SrcFrmRate = 60,
--		fr32DstFrmRate = (chanId == 1 and 60 or 30),
--		u32BitRate = (chanId == 1 and 4000000 or 1000000),
--#		u32FluctuateLevel = 0
--
-- For VBR:
--		u32Gop = 60,
--#		u32StatTime = 1,
--#		u32SrcFrmRate = 60,
--		fr32DstFrmRate = (chanId == 1 and 60 or 30),
--		u32MaxBitRate = (chanId == 1 and 4000000 or 1000000),
--		u32MinQp = 10,
--		u32MaxQp = 48

function channel_meta:SetBitrateParam( brMode, param )
	local changed = self._FORCE_UPDATE
	if brMode then
		brMode = brMode:upper()
		if brMode ~= "CBR" and brMode ~= "VBR" then
			_dbg.Error( "ENCODER", "Invalid bitrate control mode of '" .. brMode .. "', only CBR/VBR supported." )
			return
		end
		if brMode ~= self.currentBrMode then
			self.currentBrMode = brMode
			changed = true
		end
	else
		brMode = self.currentBrMode
	end

	if type(param) == "table" then
		local k, v
		for k,v in pairs(param) do
			if v ~= self.bitrateOpts[brMode].attr[ k ] then
				self.bitrateOpts[brMode].attr[k] = v
				changed = true
			end
		end
	end

	if changed and self.vencChannel then
		self:UpdateEncoding(true) --on-fly mode.
	end
end

function channel_meta:RequestIDR(now)
	if self.vencChannel then
		self.vencChannel:RequestIDR(now)
	end
end

function channel_meta:BindSourceChannel(srcChan)
	if not srcChan or srcChan == self.srcChannel then
		return
	end

	if self.initialized then
		self.vencChannel:Stop()
		self.srcChannel:Disable()
		SYS.UnBind( self.srcChannel, self.vencChannel )
	end

	self.srcChannel = srcChan

	if self.initialized then
		r, errmsg = SYS.Bind( self.srcChannel, self.vencChannel )
		if not r then
			self.srcChannel = nil
			_dbg.Error( "ENCODER", "Fail to bind video source channel to encoder, for encoding channel " .. self.id .. ", reason:", errmsg )
			return
		end

		if self.started then
			self.srcChannel:Enable()
			self.vencChannel:Start()
		end
	end
end

channel_meta.__index = channel_meta

--------------------------------------------------------------

-- Create an encoder channel.
-- Return an encoder channel object.
-- [srcChan] optional argument specify the binding video source channel, you can also
-- call channel:BindSourceChannel(...) after creating this channel.
function _M.CreateChannel(chanId, srcChan)
	chanId = chanId or 1
	if chanId < 1 or chanId > _M.MAX_CHANNELS then
		_dbg.Error( "ENCODER", "Invalid encoding channel ID passed. It's only support 1-" .. _M.MAX_CHANNELS .. " channels." )
		return nil
	end

	if _M.channels[chanId] then
		_dbg.Info( "ENCODER", "The requesting channel (ID=" .. chanId .. ") is already created, returns the exist channel object." )
		return _M.channels[chanId]
	end

	local chanObj = {
		id = chanId,
		srcChannel = srcChan,
		vencChannel = nil,

		srcImg = { width = 1920, height = 1080, frameRate = 60 },
		dstImg = { 
			width = 0,
			height = 0,
			frameRate = 0
		}, -- 0 means keep same with srcImg.

		encodingOpt = {
			type = "H264",
			attr = {
				u32MaxPicWidth = 1920,
				u32MaxPicHeight = 1080,
				u32BufSize = 1920*1080*2,
				u32Profile = 2, --Profile: 0: Baseline, 1: MP, 2: HP
				u32PicWidth = 1920,
				u32PicHeight = 1080,
				u32BFrameNum = 0,
				bByFrame = true,
				u32RefNum = 0
			}
		},

		bitrateOpts = {
			["CBR"] = {
				mode = "H264CBR",
				attr = {
					u32Gop = 60,
					u32StatTime = 1,
					u32SrcFrmRate = 30,
					fr32DstFrmRate = 30,
					u32BitRate = 4000,
					u32FluctuateLevel = 1
				}
			},

			["VBR"] = {
				mode = "H264VBR",
				attr = {
					u32Gop = 60,
					u32StatTime = 1,
					u32SrcFrmRate = 30,
					fr32DstFrmRate = 30,
					u32MaxBitRate = 4000,
					u32MinQp = 10,
					u32MaxQp = 48
				}
			}
		},

		currentBrMode = "CBR"
	}

	local r = setmetatable( chanObj, channel_meta )
	_M.channels[ chanId ] = r
	return r
end

------------------------

return _M

