local coaf = require "coaf.core"
local _dbg = require "coaf.debug"

local hiperiphery = require "hiperiphery"
local HIMM = hiperiphery.HIMM
local I2C = hiperiphery.I2C

local mo_font = require "mod_Font"
local bit = require("bit32")
local mo_Hanz = require "mod_HanzIcon"
local pt = require "pl.pretty"

-- SH1106

local _M = {
	i2c_dev = "/dev/i2c-0",
	BASE_ADDR = 0x78,
    i2c_handle = nil,
    Max_Column = 128,
    Max_Row	= 64,
    Brightness = 0xFF,
    X_WIDTH = 128,
    Y_WIDTH = 64,	
    CMD = {
        -- Constants 
        SSD1306_SETCONTRAST = 0x81,
        SSD1306_DISPLAYALLON_RESUME = 0xA4,
        SSD1306_DISPLAYALLON = 0xA5,
        SSD1306_NORMALDISPLAY = 0xA6,
        SSD1306_INVERTDISPLAY = 0xA7,
        SSD1306_DISPLAYOFF = 0xAE,
        SSD1306_DISPLAYON = 0xAF,
        SSD1306_SETDISPLAYOFFSET = 0xD3,
        SSD1306_SETCOMPINS = 0xDA,
        SSD1306_SETVCOMDETECT = 0xDB,
        SSD1306_SETDISPLAYCLOCKDIV = 0xD5,
        SSD1306_SETPRECHARGE = 0xD9,
        SSD1306_SETMULTIPLEX = 0xA8,
        SSD1306_SETLOWCOLUMN = 0x00,
        SSD1306_SETHIGHCOLUMN = 0x10,
        SSD1306_SETSTARTLINE = 0x40,
        SSD1306_MEMORYMODE = 0x20,
        SSD1306_COLUMNADDR = 0x21,
        SSD1306_PAGEADDR = 0x22,
        SSD1306_COMSCANINC = 0xC0,
        SSD1306_COMSCANDEC = 0xC8,
        SSD1306_SEGREMAP = 0xA0,
        SSD1306_CHARGEPUMP = 0x8D,
        SSD1306_EXTERNALVCC = 0x1,
        SSD1306_SWITCHCAPVCC = 0x2,
    
        -- Scrolling constants
        SSD1306_ACTIVATE_SCROLL = 0x2F,
        SSD1306_DEACTIVATE_SCROLL = 0x2E,
        SSD1306_SET_VERTICAL_SCROLL_AREA = 0xA3,
        SSD1306_RIGHT_HORIZONTAL_SCROLL = 0x26,
        SSD1306_LEFT_HORIZONTAL_SCROLL = 0x27,
        SSD1306_VERTICAL_AND_RIGHT_HORIZONTAL_SCROLL = 0x29,
        SSD1306_VERTICAL_AND_LEFT_HORIZONTAL_SCROLL = 0x2A
    }
}

function _M.i2c_Init()
	local done,err = pcall( function() _M.i2c_handle = I2C( _M.i2c_dev ) end )
	if not done then
		if type(err) == "table" then
			_dbg.Error( "OLED_CONTROL", "Could not open I2C device for OLED controlling. ErrMsg:", err.message or "---" )
		else
			_dbg.Error( "OLED_CONTROL", "Could not open I2C device for OLED controlling. ErrMsg:", err )
        end
	else
		_dbg.Info( "OLED_CONTROL", "OLED controller openned, at " .. string.format("%02x",_M.BASE_ADDR) .. "@" .. _M.i2c_dev )
	end
end

function _M.WriteReg( addr, itemArr )
	if not _M.i2c_handle then
		return false, "Invalid I2C handle"
	end
    if not itemArr or type(itemArr) ~= "table" or #itemArr < 1 then
        _dbg.Info( "OLED_CONTROL", "Invalid data" )
		return false, "Invalid data"
	end

	--table.insert( itemArr, 1, addr )
	local addrs = {}
	local i
	for i=1, #itemArr do
		addrs[i] = addr+i-1
	end
    local msg = { addrs, itemArr }
    
    local done,err = pcall( function() _M.i2c_handle:transfer( _M.BASE_ADDR>>1, msg ) end )
    -- _dbg.Info( "OLED_CONTROL", "transfer i2c data done",done,err )
    -- _dbg.Info( "OLED_CONTROL", "transfer i2c data err",done,err )
    if not done then        
		if type(err) == "table" then
			print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err.message or "---" )
			return nil, err.message or "---"
		else
			print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err )
			return nil, err
		end
	else
		return true
	end
end



function _M.SendCMD(cmd)
    local control = 0x00
    _M.WriteReg( control, { cmd } )	
end


function _M.SendByte(data)
    local control = 0x40
    _M.WriteReg( control, {data} )	    
end

function _M.OLED_Init()
    _M.i2c_Init()                                            -- init i2c

    _M.SendCMD(_M.CMD.SSD1306_DISPLAYOFF)                    -- 0xAE

    -- 设置显示时钟分频因子/振荡器频率
    _M.SendCMD(_M.CMD.SSD1306_SETDISPLAYCLOCKDIV)            -- 0xD5
    _M.SendCMD(0x80)                                       -- the suggested ratio 0x80
    _M.SendCMD(_M.CMD.SSD1306_SETMULTIPLEX)                  -- 0xA8
    _M.SendCMD(0x3F)

    -- 设置显示抵消移位映射内存计数器 
    _M.SendCMD(_M.CMD.SSD1306_SETDISPLAYOFFSET)              -- 0xD3
    _M.SendCMD(0x0)                                        -- no offset


    -- 电荷泵设置（初始化时必须打开，否则看不到显示
    _M.SendCMD(_M.CMD.SSD1306_SETSTARTLINE | 0x0)            -- line0
    _M.SendCMD(_M.CMD.SSD1306_CHARGEPUMP)                    -- 0x8D
    -- if self._vccstate == SSD1306_EXTERNALVCC:
    --     _M.SendCMD(0x10)
    -- else:
    --     _M.SendCMD(0x14)
    _M.SendCMD(0x14)

    -- 设置内存寻址方式   00: 表示水平寻址方式 01: 表示垂直寻址方式 10: 表示页寻址方式（默认方式）
    _M.SendCMD(_M.CMD.SSD1306_MEMORYMODE)                    -- 0x20
    _M.SendCMD(0x02)                                        -- 0x0 act like ks0108
    _M.SendCMD(_M.CMD.SSD1306_SEGREMAP | 0x1)
    _M.SendCMD(_M.CMD.SSD1306_COMSCANDEC)
    _M.SendCMD(_M.CMD.SSD1306_SETCOMPINS)                    -- 0xDA
    _M.SendCMD(0x12)
    _M.SendCMD(_M.CMD.SSD1306_SETCONTRAST)                   -- 0x81
    -- if self._vccstate == SSD1306_EXTERNALVCC:
    --     _M.SendCMD(0x9F)
    -- else:
    --     _M.SendCMD(0xCF)
    _M.SendCMD(0xCF)

    -- 时钟预充电周期
    _M.SendCMD(_M.CMD.SSD1306_SETPRECHARGE)                  -- 0xd9
    -- if self._vccstate == SSD1306_EXTERNALVCC:
    --     _M.SendCMD(0x22)
    -- else:
    --     _M.SendCMD(0xF1)
    _M.SendCMD(0xF1)

    _M.SendCMD(_M.CMD.SSD1306_SETVCOMDETECT)                 -- 0xDB
    _M.SendCMD(0x40)
    _M.SendCMD(_M.CMD.SSD1306_DISPLAYALLON_RESUME)           -- 0xA4
    _M.SendCMD(_M.CMD.SSD1306_NORMALDISPLAY)                 -- 0xA6
    _M.OLED_Set_Pos( 0,0 )
end

function _M.OLED_Display_Off()
    _M.SendCMD(0x8D, OLED_Order)  
    _M.SendCMD(0x10, OLED_Order)   
    _M.SendCMD(0xAE, OLED_Order)      
end

function _M.OLED_Display_On()
    _M.SendCMD(0x8D, OLED_Order)  
    _M.SendCMD(0x14, OLED_Order) -- bit2   0：关闭        1：打开
    _M.SendCMD(0xAF, OLED_Order) -- 0xAF:开显示     
end

function _M.OLED_FillPicture( data)
    local j = 0
    local x,y

    for y= 0,7 do
        _M.SendCMD(0xb0+y) 
        _M.SendCMD(0x02) 
        _M.SendCMD(0x10)

        for x=0,127 do
            j = j +1
            _M.SendByte(data[j])
        end
    end
end

function _M.OLED_Clear()
    local i,x
    for i=0,7 do
        _M.SendCMD(0xb0+i) 
        _M.SendCMD(0x02) 
        _M.SendCMD(0x10) 
        for x=0,127 do
            _M.SendByte(0x00)
        end
    end
end

function _M.OLED_ClearLine(line)
    if not line then
        line = 0
    end
    _M.SendCMD(0xb0+line)
    _M.SendCMD(0x02) 
    _M.SendCMD(0x10) 
    local x    
    for x=0,127 do
        _M.SendByte(0x00)
    end
end

function _M.OLED_ClearAera(line,x0,x1)
    if not line then
        line = 0
    end
    _M.SendCMD(0xb0+line)
    _M.SendCMD((((x0+2)&0xf0)>>4)|0x10)
    _M.SendCMD(((x0+2)&0x0f))
    local x    
    for x=1,x1-x0 do
        _M.SendByte(0x00)
    end
end

function _M.OLED_Set_Pos(x, y)
	_M.SendCMD(0xb0+y)
    _M.SendCMD((((x+2)&0xf0)>>4)|0x10)
    _M.SendCMD(((x+2)&0x0f))
end 	  

function _M.OLED_ShowChar( x, y, chr, chr_size)  -- chr_size:16或者8
    local c=0
    local i=0
    c = string.byte(chr) - string.byte(' ')   -- 获得偏移 空格


    if ( x > _M.Max_Column-1) then
        x=0
        x=x+2
    end

    if (chr_size == 16) then
        _M.OLED_Set_Pos( x, y )
        for i=1,8 do
            _M.SendByte(mo_font.F8X16[c*16+i])
        end
        _M.OLED_Set_Pos( x, y+1 )
        for i=1,8 do
            _M.SendByte(mo_font.F8X16[c*16+i+8])
        end
    else
        _M.OLED_Set_Pos( x, y )
        for i=1,6 do
            _M.SendByte(mo_font.F6x8[c*6+i])
        end
    end    
end

function _M.OLED_ShowString( x, y, str, chr_size )
	for i=1,#str do
        _M.OLED_ShowChar(x,y,string.sub(str,i,i),chr_size)
        if (chr_size == 16) then
            x = x + 8
        else
            x = x + 6
        end
		if (x>120) then
            x=0
            y=y+2
        end    
    end
end






-- _M.SendCMD(_M.CMD.SSD1306_DISPLAYOFF)                   -- 关屏


-- dofile("./page.lua")

-- 显示背景图
-- _M.OLED_DisplayImg(_Page1.NETWORK)
-- _M.OLED_Clear()
-- _M.OLED_DisplayImg(_Page1.NETWORK)

function _M.OLED_DrawBMP(x0, y0, width, hight, data, negative)
    local x,y,y1
    local i=0
    if (not negative) then 
        negative = false
    end

    if (hight%8==0) then 
        y1=hight/8
    else
        y1=hight/8+1
    end

    for y = y0,y0+y1 do
        _M.OLED_Set_Pos(x0,y)       
        for x=0,width-1 do
            i=i+1
            -- 只传送指定大小数据
            if(i > #data) then
                break
            end
            
            if negative then
                _M.SendByte(bit.bnot(data[i]))
            else
                _M.SendByte(data[i])
            end
        end
    end
end

-- test()

function isHanz(s)
    local l = #string.gsub(s, "[^\128-\191]", "") 
    return (l ~= 0)
end

function StringToTable(s)
    local tb = {}
    
    --[[
    UTF8的编码规则：
    1. 字符的第一个字节范围： 0x00—0x7F(0-127),或者 0xC2—0xF4(194-244); UTF8 是兼容 ascii 的，所以 0~127 就和 ascii 完全一致
    2. 0xC0, 0xC1,0xF5—0xFF(192, 193 和 245-255)不会出现在UTF8编码中 
    3. 0x80—0xBF(128-191)只会出现在第二个及随后的编码中(针对多字节编码，如汉字) 
    ]]
    for utfChar in string.gmatch(s, "[%z\1-\127\194-\244][\128-\191]*") do
        table.insert(tb, utfChar)
    end    
    return tb
end

function _M.OLED_ShowHanz( x, y, str, chr_size )
    local stringtab = StringToTable(str)
    local i
    for i=1,#stringtab do
        if isHanz(stringtab[i]) then
            if (chr_size == 16 or chr_size == 12) and mo_Hanz['HZ'..chr_size][stringtab[i]] then
                -- _dbg.Info("OT",stringtab[i])
                _M.OLED_DrawBMP(x, y, chr_size,chr_size, mo_Hanz['HZ'..chr_size][stringtab[i]])
                if (chr_size == 16) then
                    x = x + 16
                else
                    x = x + 12
                end
            end
        else
            _M.OLED_ShowChar( x, y, stringtab[i], chr_size)
            if (chr_size == 16) then
                x = x + 8
            else
                x = x + 6
            end
        end
        
		if (x>120) then
            x=0
            y=y+2
        end    
    end
end

function test()
    _M.OLED_Init()
    _M.OLED_Clear()
    _M.SendCMD(_M.CMD.SSD1306_DISPLAYON)                     -- 亮屏
    -- _M.OLED_ShowChar(64,4,'d',16)
    -- local icon = {0xF8,0x7C,0x7E,0xFF,0xFF,0xFF,0xFF,0x7F,0xBF,0xDE,0xEC,0xF8,0x01,0x03,0x06,0x0D,0x0B,0x0D,0x0E,0x0F,0x0F,0x07,0x03,0x01}
    -- -- 画Icon
    -- _M.OLED_DrawBMP(100,4,12,12,icon)
    _M.OLED_ShowHanz(0,0,'磁盘abcAbc',16)
end
-- test()

return _M




