local _M = {}

local lfs = require "lfs"
local presetStoDir = "/data/presets"
local pretty = require "pl.pretty"

local function file_exist(path)
	local f = io.open(path, "r")
	if f then
		f:close()
		return true
	else
		return false
	end
end

--[[
description: 生成预设列表
param  {savePos:路径地址}
return {数量,table}
--]]
local function OpenPresetDB( spec_id, savePos )
	savePos = savePos or presetStoDir
	local attrs = lfs.attributes( savePos )
	if not attrs or attrs.mode ~= "directory" then
		return nil, nil
	end

	local file = io.open("/data/configs/presetApply.conf", "r")
	local presetApply
	if file then
		presetApply = file:read()
		file:close()
	end

	local presetTable = {}
	local counter = 0

	local item
	for item in lfs.dir( savePos ) do
		if item ~= "." and item ~= ".." and (not spec_id or spec_id == item) then
			attrs = lfs.attributes( savePos .. "/" .. item )
			if attrs and attrs.mode == "directory" then
				local cfgFile = savePos .. "/" .. item .. "/cfg.info"
				local x, ret = pcall( dofile, cfgFile )
				if x and type(ret) == "table" then
					counter=counter + 1
					ret.id = item
					ret.status = presetApply == item and 1 or 0
					table.insert( presetTable, ret )
				end
			end
		end
	end

	table.sort(presetTable, function(a,b)
		if not a or not b or a==b then
			return false
		else
			return a.name < b.name
		end
	end)

	return counter, presetTable
end

--------------------
--[[
description: 获取预设列表
param  {无}
return {nil or table}
--]]
function _M.GetPresetList()
	local counter, presetTable = OpenPresetDB()
	if not counter or not presetTable then
		return nil
	else
		return presetTable
	end
end

--[[
description: 标记预设应用
param  {id:id}
return {true}
--]]
function _M.MarkPreset(id)
	local file = io.open("/data/configs/presetApply.conf", "w")
	if file then
		file:write(tostring(id))
		file:close()
	end
	return true
end


--------------------

--[[
description: 添加预设
param  {name：预设配置名称，desc：描述信息，saveTo：文件路径地址}
return {id or nil}
--]]
function _M.AddPreset( name, desc, saveTo)
	local id
	local f = io.popen( "/usr/bin/uuidgen", "r" )
	if f then
		id = f:read("l")
		f:close()
	end

	if not id then
		print( "ERROR: Fail to generate preset unique id." )
		return nil, "uuidgen"
	end

	saveTo = saveTo or presetStoDir

	--Avoid name conflict.
	if name then
		local counter, items = OpenPresetDB(nil, saveTo)
		local instead_name = name
		local instead_cc = 1
		if counter and #items > 0 then
			for _i, item in ipairs(items) do
				if item.name == instead_name then
					instead_name = name .. "_" .. instead_cc
					instead_cc = instead_cc + 1
				end
			end
			name = instead_name
		end
	end

	os.execute("/bin/rm -Rf " .. saveTo .. "/" .. id )
	os.execute( "/bin/mkdir -p " .. saveTo .. "/" .. id )
	local cfgFile = io.open( saveTo .. "/" .. id .. "/cfg.info", "w" )
	if not cfgFile then
		print( "ERROR: Fail to create the preset configure file!\n" )
		lfs.rmdir( saveTo .. "/" .. id )
		return nil, "create"
	end

	os.execute( "/bin/mkdir -p " .. saveTo .. "/" .. id .. "/configs" )

	local saveOverlay = false
	local patch_cmds = {}
	local touch_files = {}

	os.execute( "/bin/cp -Rf /data/configs/*.conf " .. saveTo .. "/" .. id .. "/configs/" )
	table.insert( touch_files, "/data/configs/*.conf")

	print("/bin/mkdir -p " .. saveTo .. "/" .. id .. "/etc/network")
	os.execute( "/bin/mkdir -p " .. saveTo .. "/" .. id .. "/etc/network" )
	os.execute( "/bin/cp -Rf /etc/network/* " .. saveTo .. "/" .. id .. "/etc/network/" )
	os.execute( "/bin/cp -Rf /etc/time.conf " .. saveTo .. "/" .. id .. "/etc/" )
	table.insert( touch_files, "/etc/time.conf")
	table.insert(patch_cmds, "/bin/chmod 600 /etc/network/*")

	os.execute( "/bin/cp -Rf /data/storage/img " .. saveTo .. "/" .. id .. "/img" )

	cfgFile:write( [==[
return {
]==] )

	cfgFile:write( 'name = [==[' .. (name or id) .. ']==],\n' )
	cfgFile:write( 'desc = [==[' .. (desc or "") .. ']==],\n' )
	cfgFile:write( 'withOverlay = ' .. (saveOverlay and "true" or "false") .. ',\n' )
	cfgFile:write( [==[
}
]==] )
	cfgFile:close()

	cfgFile = io.open(saveTo .. "/" .. id .. "/prepare.cmd", "w")
	if cfgFile then
		cfgFile:write("#!/bin/sh\n")
		for _x, file in ipairs(touch_files) do
			cfgFile:write("/bin/rm -Rf " .. file .. "\n")
		end
		cfgFile:close()
	end

	cfgFile = io.open(saveTo .. "/" .. id .. "/patch.cmd", "w")
	if cfgFile then
		cfgFile:write("#!/bin/sh\n")
		for _x, cmd in ipairs(patch_cmds) do
			cfgFile:write(cmd)
			cfgFile:write("\n")
		end
		cfgFile:close()
	end

	return id
end

--[[
description: 删除预设
param  {id：文件名,savePos：文件路径地址}
return {true or false}
--]]
function _M.DeletePreset( id, savePos )
	savePos = savePos or presetStoDir
	local attrs = lfs.attributes( savePos .. "/" .. id )
	if attrs and attrs.mode == "directory" then
		os.execute( "/bin/rm -Rf " .. savePos .. "/" .. id )
		return true
	else
		return false, "Invalid preset"
	end
end

--[[
description: 导出预设
param  {id：id,savePos：路径地址}
return {true,文件地址}
--]]
function _M.ExportPreset( id, savePos )
	savePos = savePos or presetStoDir
	local counter, presetTable = OpenPresetDB(id, savePos)
	if not counter or not presetTable or #presetTable == 0 then
		return false, "Preset is not existed"
	end

	local presetName, presetItem
	presetItem = presetTable[1]
	presetName = presetItem.name or presetItem.id

	presetName = presetName:gsub("%s+", "")
	presetName = presetName:gsub("%p+", "")

	os.execute("/bin/rm -Rf /tmp/preset/exports/" .. id)
	os.execute("/bin/mkdir -p /tmp/preset/exports/" .. id)
	--Yes sure, '/tmp/preset/exports/<id>/<id>'
	os.execute("/bin/cp -Rf " .. savePos .."/" .. id .. " /tmp/preset/exports/" .. id .. "/" )

	os.execute("/bin/rm -Rf /tmp/preset/" .. presetName .. ".tar.gz")

	local _ = os.execute(  "cd /tmp/preset/exports/" .. id .. "; /bin/tar -zcvf /tmp/preset/" .. presetName .. ".tar.gz " .. id .."; /bin/chmod 644 /tmp/preset/" .. presetName .. ".tar.gz" )
	os.execute("/bin/rm -Rf /tmp/preset/exports/" .. id)
	return "/preset/" .. presetName .. ".tar.gz"
end

--[[
description: 导入预设
param  {filepath：文件路径地址}
return {true}
--]]
function _M.ImportPreset( filepath )
	local f = io.open(filepath, "r")
	if not f then
		return  false, "Invalid preset file"
	end

	local tmppath = "/tmp/preset/imports"
	local is_ok = os.execute("/bin/mkdir -p " .. tmppath)



	if not is_ok then
		os.execute("/bin/rm " .. filepath)
		return false, "Import preset failed"
	end

	--os.execute("/bin/rm -Rf " .. tmppath .. "/*" )
	-- is_ok = os.execute("/usr/bin/unzip '" .. filepath .. "' -d " .. tmppath)
	is_ok = os.execute("/bin/tar -zxvf " .. filepath .. " -C " .. tmppath)

	os.execute("/bin/rm " .. filepath)

	if not is_ok then
		os.execute("/bin/rm -Rf " .. tmppath .. "/*")
		return false, "Extract preset file failed"
	end

	local id
	for item in lfs.dir( tmppath ) do
		if item ~= "." and item ~= ".." then
			local attrs = lfs.attributes(tmppath .. "/" .. item)
			if attrs and attrs.mode == "directory" then
				id = item
				break
			end
		end
	end

	if not id then
		os.execute("/bin/rm -Rf " .. tmppath .. "/*")
		return false, "Bad preset format"
	end

	os.execute("/bin/rm -Rf " .. presetStoDir .. "/" .. id)
	os.execute("/bin/cp -Rf " .. tmppath .. "/" .. id .. " " .. presetStoDir)
	os.execute("/bin/rm -Rf " .. tmppath .. "/*")

	return true, id

end

--[[
description: 加载应用预设
param  {id:id,savePos：路径地址}
return {true}
--]]
function _M.LoadPreset( id, savePos )
	savePos = savePos or presetStoDir
	local attrs = lfs.attributes(savePos .. "/" .. id)
	if not attrs or attrs.mode ~= "directory" then
		return false, "Preset is not existed"
	end

	local cfgFile = savePos .. "/" .. id .. "/cfg.info"
	local x, ret = pcall( dofile, cfgFile )

	if not x or type(ret) ~= "table" then
		return false, "Bad preset format"
	else
		ret.id = id
	end

	if file_exist(savePos .. "/" .. id .. "/prepare.cmd") then
		os.execute( "source " .. savePos .. "/" .. id .. "/prepare.cmd")
	end

	os.execute("/bin/cp -Rf " .. savePos .. "/" .. id .. "/configs/* /data/configs/")
	os.execute("/bin/cp -Rf " .. savePos .. "/" .. id .. "/etc/* /etc/")

	os.execute("/bin/rm  -Rf /data/storage/img")
	os.execute("/bin/cp -Rf " .. savePos .. "/" .. id .. "/img /data/storage/")

	if file_exist(savePos .. "/" .. id .. "/patch.cmd") then
		os.execute( "source " .. savePos .. "/" .. id .. "/patch.cmd")
	end
	_M.MarkPreset(id)
	return true, id
end

--[[
description: 修改预设
param  {id：id, name:预设配置名称, desc：预设描述}
return {true or nil}
--]]
function _M.ModifyPreset(id, name, desc)
	local f = io.open( presetStoDir .. "/" .. id .. "/cfg.info", "r" )
	if not f then
		print( "ERROR: Fail to find the preset configure file!\n" )
		lfs.rmdir( presetStoDir .. "/" .. id )
		return nil
	end

	if tostring(name) == "nil" or tostring(desc) == "nil" then
		print("ERROR: Invalid name or desc! \n")
		return nil
	end

	local all = f:read("*a")
	local new_name = all:gsub("name %= %[%=%=%[(.-)%]%=%=%]", "name = [==[" .. name .. "]==]", 1)
	local new_desc = new_name:gsub("desc %= %[%=%=%[(.-)%]%=%=%]", "desc = [==[" .. desc .. "]==]", 1)
	f:close()

	f = io.open( presetStoDir .. "/" .. id .. "/cfg.info", "w" )
	if not f then
		print("ERROR: Fail to 'w' open preset configure file!\n")
		return nil
	end

	f:write(new_desc)
	f:close()
	return true
end

return _M
