#!/usr/bin/env lua
local coaf = require "coaf.core"
local _dbg = require "coaf.debug"

local curdir = coaf.GetCwd()
coaf.ChDir(coaf.GetAppDir(arg))

_G.print = function(...)
	_dbg.Print( "FPGA-PROG", ... )
end

local peri = require "hiperiphery"
local HIMM = peri.HIMM
local m25p80 = require "hi_m25p80"

--Change back to current dir.
coaf.ChDir(curdir)

m25p80.sleep = coaf.Sleep

--Advanced: so you can configure the FPGA flash pins.
local _done, cfg_pinmux = pcall( require, "system.fpga_pinmux" )
local pinmux_ctrl
if not _done or type(cfg_pinmux) ~= "table" then
	pinmux_ctrl = {
		sw1 = { muxreg=0x120f013c, pingroup=0x121c0000, pin=3, muxval=0, default=1 },	--GPIO7_3: 0 - SPI to flash; 1 - FPGA to flash;
		sw2 = { muxreg=0x120f0138, pingroup=0x121c0000, pin=2, muxval=0, default=0 },	--GPIO7_2: 0 - FPGA to flash; 1 - SPI to flash;
		clk = { muxreg=0x120f014c, pingroup=0x12150000, pin=1, muxval=0, default=0 },  	--SPI CLK;
		din = { muxreg=0x120f0150, pingroup=0x12150000, pin=2, muxval=0, default=0 },  --SPI DIN;(Aka Flash DOUT)
		dout = { muxreg=0x120f0154, pingroup=0x12150000, pin=3, muxval=0, default=0 },  --SPI DOUT;(Aka Flash DIN)
		cs = { muxreg=0x120f0158, pingroup=0x12150000, pin=4, muxval=0, default=1 },  	--SPI CSN
		reset = { muxreg=0x120f015c, pingroup=0x12150000, pin=6, muxval=0, default=1 },  --FPGA ProgramMN ctrl.
	}
else
	pinmux_ctrl = cfg_pinmux
end

local function InitPinmux()
	local seqs = { "sw1", "sw2", "clk", "din", "dout", "cs", "reset" }
	local i
	local val
	for i=1, #seqs do
		if type(pinmux_ctrl[ seqs[i] ]) == "table" then
		--Only if configured.
			--NOTE: Sometimes, the PIN has no delcared mux control register.
			if pinmux_ctrl[ seqs[i] ].muxreg and pinmux_ctrl[ seqs[i] ].muxreg ~= 0 then
				HIMM.write( pinmux_ctrl[ seqs[i] ].muxreg, pinmux_ctrl[ seqs[i] ].muxval )
			end

			val = HIMM.read( pinmux_ctrl[ seqs[i] ].pingroup + 0x400 )
			if seqs[i] == "din" then
				val = val & ~(1<<pinmux_ctrl[ seqs[i] ].pin)
			else
				val = val | (1<<pinmux_ctrl[ seqs[i] ].pin)
			end
			HIMM.write( pinmux_ctrl[ seqs[i] ].pingroup + 0x400, val )
			if seqs[i] ~= "din" then
				HIMM.write( pinmux_ctrl[ seqs[i] ].pingroup + (1<<(pinmux_ctrl[ seqs[i] ].pin+2)), pinmux_ctrl[ seqs[i] ].default << pinmux_ctrl[ seqs[i] ].pin )
			end
		--End.
		else
			pinmux_ctrl[ seqs[i] ] = nil
		end
	end
end

local verify = false
local download = false
local downsize = "1M"

local file = nil
if arg[1] == "--help" then
	_dbg.Print( "FPGA-PROG", "Usage: program_fpga.lua [--verify|-v] [--download|-d <size>] <filename>" )
	os.exit(0)
elseif arg[1] == "-v" or arg[1] == "--verify" then
	verify = true
	file = arg[2]
elseif arg[1] == '-d' or arg[1] == "--download" then
	download = true
	downsize = arg[2]
	file = arg[3]
else
	file = arg[1]
end

if not file then
	_dbg.Print( "FPGA-PROG", "Usage: program_fpga.lua [--verify|-v] [--download|-d <size>] <filename>" )
	os.exit(0)
end

coaf.Init("progfpga")

local function Main(args)
	_dbg.Info( "FPGA-PROG", "Initialize pinmux..." )
	InitPinmux()
	if not pinmux_ctrl.clk or not pinmux_ctrl.cs or not pinmux_ctrl.din or not pinmux_ctrl.dout then
		_dbg.Error( "FPGA-PROG", "Invalid FPGA flash SPI pins configured!" )
		os.exit(1)
	end

	local pinOpts = "clk=" .. string.format("0x%08x",pinmux_ctrl.clk.pingroup) .. ":" .. pinmux_ctrl.clk.pin .. "," ..
					"cs=" .. string.format("0x%08x",pinmux_ctrl.cs.pingroup) .. ":" .. pinmux_ctrl.cs.pin .. "," ..
					"din=" .. string.format("0x%08x",pinmux_ctrl.din.pingroup) .. ":" .. pinmux_ctrl.din.pin .. "," ..
					"dout=" .. string.format("0x%08x",pinmux_ctrl.dout.pingroup) .. ":" .. pinmux_ctrl.dout.pin
	
	_dbg.Info( "FPGA-PROG", "Create M25P80 chip instance...", pinOpts )
	local chip = m25p80.createNew( pinOpts )
	if not chip then
		_dbg.Error( "FPGA-PROG", "Fail to create M25P80 chip instance" )
		os.exit(1)
	end
	
	_dbg.Info( "FPGA-PROG", "Switch FPGA flash to CPU for programming..." )
	if pinmux_ctrl.sw1 then
		HIMM.write( pinmux_ctrl.sw1.pingroup + (1<<(pinmux_ctrl.sw1.pin+2)), 0 << pinmux_ctrl.sw1.pin )
	end
	if pinmux_ctrl.sw2 then
		HIMM.write( pinmux_ctrl.sw2.pingroup + (1<<(pinmux_ctrl.sw2.pin+2)), 1 << pinmux_ctrl.sw2.pin )
	end
	_dbg.Info( "FPGA-PROG", "Check M25P80 chip's UUID..." )
	--coaf.Sleep(3)
	local uuid = chip:GetUUID()
	if not uuid then
		_dbg.Info( "FPGA-PROG", "Switch FPGA flash back to FPGA." )
		if pinmux_ctrl.sw1 then
			HIMM.write( pinmux_ctrl.sw1.pingroup + (1<<(pinmux_ctrl.sw1.pin+2)), 1 << pinmux_ctrl.sw1.pin )
		end
		if pinmux_ctrl.sw2 then
			HIMM.write( pinmux_ctrl.sw2.pingroup + (1<<(pinmux_ctrl.sw2.pin+2)), 0 << pinmux_ctrl.sw2.pin )
		end
		_dbg.Error( "FPGA-PROG", "Fail to get chip UUID!" )
		os.exit(1)
	else
		_dbg.Info( "FPGA-PROG", "Got M25P80 UUID:", uuid )
	end
	
	if verify then
		_dbg.Info( "FPGA-PROG", "Verify chip compares to file:", file, "..." )
		if chip:VerifyFile(file) then
			_dbg.Info( "FPGA-PROG", "Chip verify DONE." )
		else
			_dbg.Error( "FPGA-PROG", "Chip verify FAIL!" )
		end
	elseif download then
		_dbg.Info( "FPGA-PROG", "Download chip content to file:", file, ", size=", downsize, "..." )
		if chip:DownloadFile(file, downsize) then
			_dbg.Info( "FPGA-PROG", "Chip download DONE." )
		else
			_dbg.Error( "FPGA-PROG", "Chip download FAIL!" )
		end
	else
		_dbg.Info( "FPGA-PROG", "Program binary file:", file, "..." )
		if chip:ProgramFile(file,true) then
			_dbg.Info( "FPGA-PROG", "Chip programmed DONE." )
		else
			_dbg.Error( "FPGA-PROG", "Chip programmed FAIL!" )
		end
	end
	
	_dbg.Info( "FPGA-PROG", "Switch FPGA flash back to FPGA." )
	if pinmux_ctrl.sw1 then
		HIMM.write( pinmux_ctrl.sw1.pingroup + (1<<(pinmux_ctrl.sw1.pin+2)), 1 << pinmux_ctrl.sw1.pin )
	end

	if pinmux_ctrl.sw2 then
		HIMM.write( pinmux_ctrl.sw2.pingroup + (1<<(pinmux_ctrl.sw2.pin+2)), 0 << pinmux_ctrl.sw2.pin )
	end
	
	if not verify and not download and pinmux_ctrl.reset then
		_dbg.Info( "FPGA-PROG", "Reset the FPGA..." )
		HIMM.write( pinmux_ctrl.reset.pingroup + (1<<(pinmux_ctrl.reset.pin+2)), 0 << pinmux_ctrl.reset.pin )
		coaf.Sleep(0.5)
		HIMM.write( pinmux_ctrl.reset.pingroup + (1<<(pinmux_ctrl.reset.pin+2)), 1 << pinmux_ctrl.reset.pin )
	end
	coaf.Exit(0)
end

coaf.Start(Main)

os.exit(coaf.CoAF_returnCode or 0)

