local coaf = require 'coaf.core'
local thread = require 'coaf.thread'
local _dbg = require 'coaf.debug'
-- local utils = require "coaf.utils"
local hiperiphery = require 'hiperiphery'
local HIMM = hiperiphery.HIMM
local I2C = hiperiphery.I2C
local bit = require 'bit32'
local pl = require 'pl.pretty'
local Date = require 'pl.Date'
local TOOLS = require 'system.cmdtools'

pcall(require, 'lsystime')

--------
local _M = {
    I2C_DEV = "scl=0x121c0000:4 sda=0x121c0000:5",
    
    PINMUX_CTRL = {
        scl = { muxreg=0x120f0140, pingroup=0x121c0000, pin=4, muxval=0 },  --GPIO7_4: SCL
        sda = { muxreg=0x120f0144, pingroup=0x121c0000, pin=5, muxval=0 },  --GPIO7_5: SDA
    },

    __pin_init = false,
    BASE_ADDR = 0xA2, -- read A3h and write A2h
    i2c_handle = nil,
    dtfmt = 'yyyy-mm-dd HH:MM:SS'
}

function _M.ReadReg( addr, itemArr )
    if not _M.i2c_handle then              
        return false, 'Invalid I2C handle'
    end 

    if not itemArr or type(itemArr) ~= "table" then
        itemArr = { 0xff, flags = I2C.I2C_M_RD }
    else
        if #itemArr < 1 then
            itemArr[1] = 0xff
        end
        if not itemArr.flags then
            itemArr.flags = I2C.I2C_M_RD
        else
            itemArr.flags = itemArr.flags | I2C.I2C_M_RD
        end
    end

    local msg = { {addr}, itemArr }
    local done,err = pcall( function() _M.i2c_handle:transfer( _M.BASE_ADDR>>1, msg ) end )
    -- _M.i2c_handle:close()
    if not done then
        if type(err) == "table" then
            --print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err.message or "---" )
            return nil, err.message or "---"
        else
            --print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err )
            return nil, err
        end
    else
        return itemArr
    end
end

-----------------------------------
function _M.WriteReg( addr, itemArr )
    if not _M.i2c_handle then              
        return false, 'Invalid I2C handle'
    end 

    if not itemArr then
        --print( "ERROR: Invalid writing data" )
        return nil, "Invalid data"
    elseif type(itemArr) ~= "table" then
        itemArr = { itemArr }
    end

    local msg = {{ addr, table.unpack(itemArr) }}
    local done,err = pcall( function() _M.i2c_handle:transfer( _M.BASE_ADDR>>1, msg ) end )
    -- _M.i2c_handle:close()
    if not done then
        if type(err) == "table" then
            --print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err.message or "---" )
            return nil, err.message or "---"
        else
            --print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err )
            return nil, err
        end
    else
        return itemArr
    end
end
--------

function decToBcd(val)
    local d = string.format('%d', math.modf(val / 10))
    local d1 = tonumber(d * 10)
    local d2 = val - d1
    return tonumber(d * 16 + d2)
end

function bcdToDec(val)
    local hl = bit.rshift(val, 4)
    local hh = bit.band(val, 0xf)
    local hr = string.format('%d%d', hl, hh)
    return string.format('%d%d', hl, hh)
end

function _M.InitPinmux()
    if _M.__pin_init then
        return
    end

    local pinmux_ctrl = _M.PINMUX_CTRL
    local seqs = { "scl", "sda" }
    local i
    local val
    for i=1, #seqs do
        if type(pinmux_ctrl[ seqs[i] ]) == "table" then
        --Only if configured.
            --NOTE: Sometimes, the PIN has no delcared mux control register.
            if pinmux_ctrl[ seqs[i] ].muxreg and pinmux_ctrl[ seqs[i] ].muxreg ~= 0 then
                HIMM.write( pinmux_ctrl[ seqs[i] ].muxreg, pinmux_ctrl[ seqs[i] ].muxval )
            end

            val = HIMM.read( pinmux_ctrl[ seqs[i] ].pingroup + 0x400 )
            --Make default as input.
            val = val & ~(1<<pinmux_ctrl[ seqs[i] ].pin)
            HIMM.write( pinmux_ctrl[ seqs[i] ].pingroup + 0x400, val )
        --End.
        else
            pinmux_ctrl[ seqs[i] ] = nil
        end
    end
    _M.__pin_init = true
end

function _M.OpenI2C(addr)
    _M.InitPinmux()

    -- _M.BASE_ADDR = addr or _M.BASE_ADDR

    local done,err = pcall( function() _M.i2c_handle = I2C( _M.I2C_DEV ) end )
    if not done then
        if type(err) == "table" then
            _dbg.Error("RTC", "ERROR: Could not open I2C device for FPGA controlling. ErrMsg:", err.message or "---" )
        else
            _dbg.Error("RTC", "ERROR: Could not open I2C device for FPGA controlling. ErrMsg:", err )
        end
        return false
    else
        _dbg.Info("RTC", "RTC controller openned, at " .. string.format("%02x",_M.BASE_ADDR) .. "@" .. _M.I2C_DEV )
        return true
    end

end

--------
function _M.Init()
    if LSYSTIME then
        LSYSTIME.SetAllowTimeDiff(100000) --100ms.
    end

    if not _M.OpenI2C() then
        _dbg.Error('RTC', 'Could not open I2C')
        os.exit(1)
    end
end
--------

function _M.GetTime()
    local wkd = {'Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday'}
    local datatime = _M.ReadReg(0x02, {0, 0, 0, 0, 0, 0, 0})
    if datatime ~= nil then
        -- pl.dump(datatime)
        local second = bcdToDec(bit.band(datatime[1], 0x7f))
        local minute = bcdToDec(bit.band(datatime[2], 0x7f))
        local hour = bcdToDec(bit.band(datatime[3], 0x3f))
        local day = bcdToDec(bit.band(datatime[4], 0x3f))
        -- local second = wkd[tonumber(bcdToDec(bit.band(datatime[5],0x7)))],
        -- local century_bit = bit.rshift( datatime[5],7)
        local weekday = bcdToDec(bit.band(datatime[5], 0x7))
        local month = bcdToDec(bit.band(datatime[6], 0x1f))
        local year = bcdToDec(datatime[7])
        if bit.band(datatime[6], 0x80) > 0 then
            year = '19' .. year
        else
            year = '20' .. year
        end

        local datetime_str = year .. '-' .. month .. '-' .. day .. ' ' .. hour .. ':' .. minute .. ':' .. second
        return datetime_str
    else
        -- _dbg.Error('RTC', 'GetTime error!!!')
        return nil
    end
end

-- print(_M.GetTime())

function _M.SetTime(datetime_str)
    if not datetime_str then
        _dbg.Error('RTC', 'no datatime string!')
        return false
    end

    local df = Date.Format(_M.dtfmt)
    local dt = df:parse(datetime_str)

    --print(df:tostring(dt))
    if dt ~= nil and dt.tab then
        local century_month = 0x0
        local year = 0
        if dt.tab.year < 2000 then
            century_month = 0x80
            year = decToBcd(dt.tab.year - 1900)
        else
            century_month = 0x00
            year = decToBcd(dt.tab.year - 2000)
        end
        century_month = bit.bor(century_month, decToBcd(dt.tab.month))
        _M.WriteReg(
            0x02,
            {
                decToBcd(dt.tab.sec),
                decToBcd(dt.tab.min),
                decToBcd(dt.tab.hour),
                decToBcd(dt.tab.day),
                decToBcd(dt.tab.wday),
                century_month,
                year
            }
        )
    else
        --  _dbg.Error('RTC', 'SetTime() parse string error!!!')
        return false,'SetTime() parse time string error!!!'
    end
    return true
end
function _M.Store()
    local datetime = os.date('%Y-%m-%d %H:%M:%S', os.time())
    _dbg.Info('RTC', 'Set UTC RTC datetime=', datetime)
    return _M.SetTime(datetime)    
end

function _M.Load()
    local datetime = _M.GetTime()
	if not datetime then                                 
        _dbg.Error('RTC', 'GetTime error!')                                  
		return false                                      
	end
    _dbg.Info('RTC', 'Get UTC RTC datetime=', datetime)
    local df = Date.Format('yyyy-mm-dd HH:MM:SS')
    local dt = df:parse(datetime)

    if LSYSTIME then
        LSYSTIME.AdjustTime(datetime, -1000000, 1000000)
    else
        os.execute(TOOLS.SET_DATETIME .. " '" .. datetime .. "'")
    end
    return true
end

function _M.Destroy()
    if _M.i2c_handle then
        _M.i2c_handle:close()
    end
end

--------
return _M
