local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local hiperiphery = require "hiperiphery"
local HIMM = hiperiphery.HIMM
local I2C = hiperiphery.I2C

local _M = {
	BASE_ADDR = 0xf2,
	BLINK_VERYFAST = 0x55555555,
	BLINK_FAST = 0x33333333,
	BLINK = 0x0f0f0f0f,
	BLINK_SLOW = 0x00ff00ff,
	ON = 0xffffffff,
	OFF = 0x00000000,
	lights = {},
	curStep = 0
}

local fpgaStatusMap = {
	[_M.BLINK_VERYFAST] = {6,6},
	[_M.BLINK_FAST] = {13,12},
	[_M.BLINK] = {25,25},
	[_M.BLINK_SLOW] = {50,50},
	[_M.ON] = {255,0},
	[_M.OFF] = {0,0}
}

function _M.Init(i2cDev)
	local done,err = pcall( function() _M.i2c_handle = I2C( i2cDev or "/dev/i2c-0" ) end )
	if not done then
		if type(err) == "table" then
			_dbg.Error( "LED", "Could not open I2C device for FPGA controlling. ErrMsg:", err.message or "---" )
		else
			_dbg.Error( "LED", "Could not open I2C device for FPGA controlling. ErrMsg:", err )
		end
	else
		_dbg.Info( "LED", "FPGA controller openned, at " .. string.format("%02x",_M.BASE_ADDR) .. "@" .. (i2cDev or "default") )
	end
	_M.curStep = 0
end

function _M.ReadReg( addr, itemArr )
	if not _M.i2c_handle then
		return nil, "Invalid I2C handle"
	end
	if not itemArr or type(itemArr) ~= "table" then
		itemArr = { 0xff, flags = I2C.I2C_M_RD }
	else
		if #itemArr < 1 then
			itemArr[1] = 0xff
		end
		if not itemArr.flags then
			itemArr.flags = I2C.I2C_M_RD
		else
			itemArr.flags = itemArr.flags | I2C.I2C_M_RD
		end
	end

	local addrs = {}
	local i
	for i=1, #itemArr do
		addrs[i] = addr+i-1
	end

	local msg = { addrs, itemArr }
	local done,err = pcall( function() _M.i2c_handle:transfer( _M.BASE_ADDR>>1, msg ) end )
	if not done then
		if type(err) == "table" then
			--print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err.message or "---" )
			return nil, err.message or "---"
		else
			--print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err )
			return nil, err
		end
	else
		return itemArr
	end
end

-----------------------------------

function _M.WriteReg( addr, itemArr )
	if not _M.i2c_handle then
		return false, "Invalid I2C handle"
	end
	if not itemArr or type(itemArr) ~= "table" or #itemArr < 1 then
		return false, "Invalid data"
	end

	--table.insert( itemArr, 1, addr )
	local addrs = {}
	local i
	for i=1, #itemArr do
		addrs[i] = addr+i-1
	end
	local msg = { addrs, itemArr }
	local done,err = pcall( function() _M.i2c_handle:transfer( _M.BASE_ADDR>>1, msg ) end )
	if not done then
		if type(err) == "table" then
			--print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err.message or "---" )
			return nil, err.message or "---"
		else
			--print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err )
			return nil, err
		end
	else
		return true
	end
end

-- **Pin options**
-- {
-- 		fpga = true,
-- 		fpga_baseAddr = regAddr,
-- 		muxreg = ...,
-- 		muxval = ...,
-- 		pingroup = ...,
-- 		pin = ...,
-- 		pol = 0/1
-- }
function _M.CreateLight(name, pinOpts)
	if _M.lights[name] then
		return true
	end
	if not pinOpts then
		return false
	end
	if not pinOpts.fpga and (not pinOpts.pingroup or not pinOpts.pin) then
		return false
	elseif pinOpts.fpga and not pinOpts.fpga_baseAddr then
		return false
	end

	if pinOpts.fpga then
		local light = {
			fpga = true,
			fpga_baseAddr = pinOpts.fpga_baseAddr,
			status = {0, 0},
			forceStatus = nil
		}
		_M.WriteReg(pinOpts.fpga_baseAddr, {0})	--Disable light.
		_M.lights[name] = light
	else
		if pinOpts.muxreg then
			HIMM.write( pinOpts.muxreg, pinOpts.muxval or 0 )
		end
		local val = HIMM.read( pinOpts.pingroup + 0x400 )
		val = val | (1<<pinOpts.pin)
		HIMM.write( pinOpts.pingroup + 0x400, val )
		local pol = pinOpts.pol or 0
		HIMM.write( pinOpts.pingroup + (1<<(pinOpts.pin+2)), (pol==0 and 1 or 0)<<pinOpts.pin )
		local light = {
			writeReg = pinOpts.pingroup + (1<<(pinOpts.pin+2)),
			onVal = pol << pinOpts.pin,
			offVal = (pol==0 and 1 or 0) << pinOpts.pin,
			status = 0,
			forceStatus = nil
		}
	
		_M.lights[name] = light
	end
	return true
end

function _M.RemoveLight(name)
	_M.lights[name] = nil
end

function _M.SetLightStatus(name, status, offAfter)
	local light = _M.lights[name]
	if light then
		if light.fpga then
			if type(status) ~= "table" then
				status = fpgaStatusMap[status] or {0,0}
			end

			if not light.forceStatus then
				if ( not status[1] or status[1] == 0 ) and ( not status[2] or status[2] == 0 ) then
					_M.WriteReg( light.fpga_baseAddr, {0} )
				else
					_M.WriteReg( light.fpga_baseAddr+1, {status[1]} )
					_M.WriteReg( light.fpga_baseAddr+2, {status[2]} )
					_M.WriteReg( light.fpga_baseAddr, {1} )
				end
			end
		end
		light.status = status
		light.offAfter = offAfter
	end
end

function _M.Force(name, status, offAfter)
	local light = _M.lights[name]
	if light then
		if light.fpga then
			if type(status) ~= "table" then
				status = fpgaStatusMap[status] or {0,0}
			end

			if ( not status[1] or status[1] == 0 ) and ( not status[2] or status[2] == 0 ) then
				_M.WriteReg( light.fpga_baseAddr, {0} )
			else
				_M.WriteReg( light.fpga_baseAddr+1, {status[1]} )
				_M.WriteReg( light.fpga_baseAddr+2, {status[2]} )
				_M.WriteReg( light.fpga_baseAddr, {1} )
			end
		end
		light.forceStatus = status
		light.forceOffAfter = offAfter
	end
end

function _M.NoForce(name)
	local light = _M.lights[name]
	if light then
		light.forceStatus = nil
		light.forceOffAfter = nil
		if light.fpga then
			local status = light.status
			if ( not status[1] or status[1] == 0 ) and ( not status[2] or status[2] == 0 ) then
				_M.WriteReg( light.fpga_baseAddr, {0} )
			else
				_M.WriteReg( light.fpga_baseAddr+1, {status[1]} )
				_M.WriteReg( light.fpga_baseAddr+2, {status[2]} )
				_M.WriteReg( light.fpga_baseAddr, {1} )
			end
		end
	end
end

function _M.Step()
	local n,light
	local status
	local statusChanged
	for n,light in pairs(_M.lights) do
		statusChanged = false
		if light.forceOffAfter then
			light.forceOffAfter = light.forceOffAfter - 1
			if light.forceOffAfter <= 0 then
				light.forceOffAfter = nil
				light.forceStatus = nil
				statusChanged = true
			end
		end

		if light.offAfter then
			light.offAfter = light.offAfter - 1
			if light.offAfter <= 1 then
				light.offAfter = nil
				if light.fpga then
					light.status = {0,0}
				else
					light.status = _M.OFF
				end
				statusChanged = true
			end
		end

		if statusChanged and light.fpga then
			status = light.status
			if ( not status[1] or status[1] == 0 ) and ( not status[2] or status[2] == 0 ) then
				_M.WriteReg( light.fpga_baseAddr, {0} )
			else
				_M.WriteReg( light.fpga_baseAddr+1, {status[1]} )
				_M.WriteReg( light.fpga_baseAddr+2, {status[2]} )
				_M.WriteReg( light.fpga_baseAddr, {1} )
			end
		end

		if not light.fpga then
			status = light.forceStatus or light.status or 0
			if ((status >> _M.curStep) & 0x01) ~= 0 then
				HIMM.write( light.writeReg, light.onVal )
			else
				HIMM.write( light.writeReg, light.offVal )
			end
		end
	end
	_M.curStep = _M.curStep + 1
	if _M.curStep >= 32 then
		_M.curStep = 0
	end
end

return _M

