local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local hiperiphery = require "hiperiphery"
local HIMM = hiperiphery.HIMM
local I2C = hiperiphery.I2C

local _M = {
	I2C_DEV = "scl=0x121c0000:4 sda=0x121c0000:5",

	PINMUX_CTRL = {
		scl = { muxreg=0x120f0140, pingroup=0x121c0000, pin=4, muxval=0 },	--GPIO7_4: SCL
		sda = { muxreg=0x120f0144, pingroup=0x121c0000, pin=5, muxval=0 },	--GPIO7_5: SDA
	},

	BASE_ADDR = 0xe2,
	SAVED_VERSION = nil,
	__pin_init = false
}


function _M.InitPinmux()
	if _M.__pin_init then
		return
	end

	local pinmux_ctrl = _M.PINMUX_CTRL
	local seqs = { "scl", "sda" }
	local i
	local val
	for i=1, #seqs do
		if type(pinmux_ctrl[ seqs[i] ]) == "table" then
		--Only if configured.
			--NOTE: Sometimes, the PIN has no delcared mux control register.
			if pinmux_ctrl[ seqs[i] ].muxreg and pinmux_ctrl[ seqs[i] ].muxreg ~= 0 then
				HIMM.write( pinmux_ctrl[ seqs[i] ].muxreg, pinmux_ctrl[ seqs[i] ].muxval )
			end

			val = HIMM.read( pinmux_ctrl[ seqs[i] ].pingroup + 0x400 )
			--Make default as input.
			val = val & ~(1<<pinmux_ctrl[ seqs[i] ].pin)
			HIMM.write( pinmux_ctrl[ seqs[i] ].pingroup + 0x400, val )
		--End.
		else
			pinmux_ctrl[ seqs[i] ] = nil
		end
	end
	_M.__pin_init = true
end

function _M.ReadReg( addr, itemArr )
	_M.InitPinmux()

	local done,err = pcall( function() _M.i2c_handle = I2C( _M.I2C_DEV ) end )
	if not done then
		if type(err) == "table" then
			_dbg.Error( "HW", "Could not open I2C device for FPGA controlling. ErrMsg:", err.message or "---" )
		else
			_dbg.Error( "HW", "Could not open I2C device for FPGA controlling. ErrMsg:", err )
		end
		return nil
	else
		_dbg.Info( "HW", "FPGA controller openned, at " .. string.format("%02x",_M.BASE_ADDR) .. "@" .. _M.I2C_DEV )
	end

	if not itemArr or type(itemArr) ~= "table" then
		itemArr = { 0xff, flags = I2C.I2C_M_RD }
	else
		if #itemArr < 1 then
			itemArr[1] = 0xff
		end
		if not itemArr.flags then
			itemArr.flags = I2C.I2C_M_RD
		else
			itemArr.flags = itemArr.flags | I2C.I2C_M_RD
		end
	end

	local msg = { {addr}, itemArr }
	local done,err = pcall( function() _M.i2c_handle:transfer( _M.BASE_ADDR>>1, msg ) end )
	_M.i2c_handle:close()
	if not done then
		if type(err) == "table" then
			--print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err.message or "---" )
			return nil, err.message or "---"
		else
			--print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err )
			return nil, err
		end
	else
		return itemArr
	end
end


function _M.WriteReg( addr, itemArr )
	_M.InitPinmux()

	local done,err = pcall( function() _M.i2c_handle = I2C( _M.I2C_DEV ) end )
	if not done then
		if type(err) == "table" then
			_dbg.Error( "HW", "Could not open I2C device for FPGA controlling. ErrMsg:", err.message or "---" )
		else
			_dbg.Error( "HW", "Could not open I2C device for FPGA controlling. ErrMsg:", err )
		end
		return nil
	else
		_dbg.Info( "HW", "FPGA controller openned, at " .. string.format("%02x",_M.BASE_ADDR) .. "@" .. _M.I2C_DEV )
	end

	if not itemArr or type(itemArr) ~= "table" or #itemArr < 1 then
		return false, "Invalid data"
	end

	table.insert( itemArr, 1, addr )
	local msg = { itemArr }
	local done,err = pcall( function() _M.i2c_handle:transfer( _M.BASE_ADDR>>1, msg ) end )
	_M.i2c_handle:close()
	if not done then
		if type(err) == "table" then
			--print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err.message or "---" )
			return nil, err.message or "---"
		else
			--print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", BASE_ADDR, addr), err )
			return nil, err
		end
	else
		return true
	end
end

-----------------------------------

function _M.GetHardwareVersion()
	if _M.SAVED_VERSION then
		return _M.SAVED_VERSION
	else
		local tmp_ver
		local read_f = io.open( "/tmp/HARDWARE_VERSION", "r" )
		if read_f then
			tmp_ver = read_f:read("*l")
			read_f:close()
		end

		if tmp_ver and tmp_ver ~= "" then
			_M.SAVED_VERSION = "2.0." .. tmp_ver
			return _M.SAVED_VERSION
		end

		local ret = _M.ReadReg( 0xa0, {0x00} )
		if not ret then
			return "2.0.00"
		else
			--To avoid GPIO reuse ...
			local tmp_record_f = io.open( "/tmp/HARDWARE_VERSION", "w" )
			if tmp_record_f then
				tmp_record_f:write( string.format( "%02X", ret[1] ) )
				tmp_record_f:close()
			end
			_M.SAVED_VERSION = "2.0." .. string.format( "%02X", ret[1] )
			return _M.SAVED_VERSION
		end
	end
end

-----------------------------------

return _M

